"""Configurable JSON serialization with Protocol-based architecture.

This module provides a flexible serialization system that supports:
- Custom type encoders via the ``type_encoders`` parameter
- Multiple backend support (msgspec, orjson, standard library)
- Thread-safe caching of serializer instances
- Integration with Litestar's type_encoders pattern

Example:
    Basic usage with default encoders::

        >>> from advanced_alchemy.utils.serialization import encode_json, decode_json
        >>> import datetime
        >>> encode_json({"date": datetime.date.today()})
        '{"date":"2025-12-17"}'

    Custom type encoders::

        >>> from decimal import Decimal
        >>> class Money:
        ...     def __init__(self, amount: Decimal):
        ...         self.amount = amount
        >>> encode_json(
        ...     {"price": Money(Decimal("19.99"))},
        ...     type_encoders={Money: lambda m: str(m.amount)}
        ... )
        '{"price":"19.99"}'
"""

import datetime
import enum
import json
import threading
from abc import ABC, abstractmethod
from collections.abc import Mapping
from dataclasses import MISSING as DATACLASS_MISSING
from dataclasses import dataclass
from dataclasses import fields as dataclass_fields
from decimal import Decimal
from functools import lru_cache
from ipaddress import IPv4Address, IPv4Interface, IPv4Network, IPv6Address, IPv6Interface, IPv6Network
from pathlib import Path, PurePath
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Callable,
    Final,
    Literal,
    Optional,
    Protocol,
    TypeVar,
    Union,
    cast,
    overload,
    runtime_checkable,
)
from uuid import UUID

from sqlalchemy import RowMapping
from typing_extensions import TypeAlias, TypeGuard

from advanced_alchemy.typing import (
    ATTRS_INSTALLED,
    CATTRS_INSTALLED,
    LITESTAR_INSTALLED,
    MSGSPEC_INSTALLED,
    NUMPY_INSTALLED,
    ORJSON_INSTALLED,
    PYDANTIC_INSTALLED,
    SQLMODEL_INSTALLED,
    UNSET,
    AttrsInstance,
    AttrsLike,
    BaseModel,
    BaseModelLike,
    DTOData,
    DTODataLike,
    FailFast,
    Struct,
    StructLike,
    attrs_nothing,
    convert,
)
from advanced_alchemy.typing import attrs_asdict as asdict
from advanced_alchemy.typing import attrs_fields as fields
from advanced_alchemy.typing import attrs_has as has
from advanced_alchemy.typing import cattrs_structure as structure
from advanced_alchemy.typing import cattrs_unstructure as unstructure
from advanced_alchemy.utils.dataclass import Empty as DataclassEmpty
from advanced_alchemy.utils.dataclass import is_dataclass_instance

if TYPE_CHECKING:
    from collections.abc import Sequence

    from sqlalchemy.engine.row import Row

    from advanced_alchemy.filters import StatementFilter
    from advanced_alchemy.repository.typing import ModelT

__all__ = (
    "ATTRS_INSTALLED",
    "CATTRS_INSTALLED",
    "DEFAULT_TYPE_ENCODERS",
    "MSGSPEC_INSTALLED",
    "PYDANTIC_INSTALLED",
    "PYDANTIC_USE_FAILFAST",
    "UNSET",
    "AttrsInstance",
    "BaseModel",
    "BulkModelDictT",
    "FilterTypeT",
    "JSONSerializer",
    "ModelDTOT",
    "ModelDictListT",
    "ModelDictT",
    "MsgspecSerializer",
    "OrjsonSerializer",
    "PydanticOrMsgspecT",
    "SchemaDumpConfig",
    "StandardLibSerializer",
    "Struct",
    "SupportedSchemaModel",
    "TypeEncodersMap",
    "asdict",
    "attrs_nothing",
    "convert",
    "convert_date_to_iso",
    "convert_datetime_to_gmt_iso",
    "decode_complex_type",
    "decode_json",
    "encode_complex_type",
    "encode_json",
    "fields",
    "get_attrs_fields",
    "get_serializer",
    "get_type_adapter",
    "has",
    "has_dict_attribute",
    "is_attrs_instance",
    "is_attrs_instance_with_field",
    "is_attrs_instance_without_field",
    "is_attrs_schema",
    "is_dataclass",
    "is_dataclass_with_field",
    "is_dataclass_without_field",
    "is_dict",
    "is_dict_with_field",
    "is_dict_without_field",
    "is_dto_data",
    "is_msgspec_struct",
    "is_msgspec_struct_with_field",
    "is_msgspec_struct_without_field",
    "is_pydantic_model",
    "is_pydantic_model_with_field",
    "is_pydantic_model_without_field",
    "is_row_mapping",
    "is_schema",
    "is_schema_or_dict",
    "is_schema_or_dict_with_field",
    "is_schema_or_dict_without_field",
    "is_schema_with_field",
    "is_schema_without_field",
    "is_sqlmodel_table_model",
    "schema_dump",
    "structure",
    "unstructure",
)

# Type aliases
TypeEncodersMap = Mapping[type, Callable[[Any], Any]]
"""Mapping of types to encoder functions for custom serialization."""

T = TypeVar("T")

PYDANTIC_USE_FAILFAST = False  # leave permanently disabled for now

FilterTypeT = TypeVar("FilterTypeT", bound="StatementFilter")
"""Type variable for filter types."""

SupportedSchemaModel: TypeAlias = Union[StructLike, BaseModelLike, AttrsLike]
"""Type alias for objects that support schema conversion methods (model_dump, asdict, etc.)."""

ModelDTOT = TypeVar("ModelDTOT", bound="Union[SupportedSchemaModel, Any]")
"""Type variable for model DTOs."""

PydanticOrMsgspecT = SupportedSchemaModel
"""Type alias for supported schema models."""

ModelDictT: TypeAlias = "Union[dict[str, Any], ModelT, SupportedSchemaModel, DTODataLike[ModelT], Any]"
"""Type alias for model dictionaries."""

ModelDictListT: TypeAlias = "Sequence[Union[dict[str, Any], ModelT, SupportedSchemaModel, Any]]"
"""Type alias for model dictionary lists."""

BulkModelDictT: TypeAlias = (
    "Union[Sequence[Union[dict[str, Any], ModelT, SupportedSchemaModel, Any]], DTODataLike[list[ModelT]]]"
)
"""Type alias for bulk model dictionaries."""


class _HasDictAttribute(Protocol):
    __dict__: dict[str, Any]


@dataclass(frozen=True)
class SchemaDumpConfig:
    """Configuration for dumping schema-like service input data."""

    exclude_unset: bool = True
    exclude_none: bool = False
    exclude_defaults: bool = False
    exclude_sentinels: bool = True


DEFAULT_SCHEMA_DUMP_CONFIG: Final = SchemaDumpConfig()


def _get_schema_dump_config(exclude_unset: bool, config: "Optional[SchemaDumpConfig]") -> SchemaDumpConfig:
    """Resolve backwards-compatible schema dump configuration."""
    if config is not None:
        return config
    if exclude_unset:
        return DEFAULT_SCHEMA_DUMP_CONFIG
    return SchemaDumpConfig(exclude_unset=False, exclude_sentinels=False)


def _get_pydantic_missing_sentinel() -> Any:
    """Return Pydantic's experimental MISSING sentinel when available."""
    if not PYDANTIC_INSTALLED:
        return None
    try:
        from pydantic.experimental.missing_sentinel import MISSING
    except ImportError:  # pragma: no cover - depends on installed Pydantic version
        return None
    return MISSING


# ============================================================================
# Helper functions
# ============================================================================


def convert_datetime_to_gmt_iso(dt: datetime.datetime) -> str:
    """Convert datetime to ISO 8601 format with UTC timezone.

    If the datetime is naive (no timezone), UTC is assumed.
    The ``+00:00`` suffix is replaced with ``Z`` for consistency.

    Args:
        dt: The datetime to convert.

    Returns:
        ISO 8601 formatted datetime string with ``Z`` suffix for UTC.

    Example:
        >>> import datetime
        >>> dt = datetime.datetime(2025, 12, 17, 10, 30, 0)
        >>> convert_datetime_to_gmt_iso(dt)
        '2025-12-17T10:30:00Z'
    """
    if not dt.tzinfo:
        dt = dt.replace(tzinfo=datetime.timezone.utc)
    return dt.isoformat().replace("+00:00", "Z")


def convert_date_to_iso(dt: datetime.date) -> str:
    """Convert date to ISO 8601 format.

    Args:
        dt: The date to convert.

    Returns:
        ISO 8601 formatted date string (YYYY-MM-DD).

    Example:
        >>> import datetime
        >>> convert_date_to_iso(datetime.date(2025, 12, 17))
        '2025-12-17'
    """
    return dt.isoformat()


@lru_cache(typed=True)
def get_type_adapter(f: "type[T]") -> Any:
    """Caches and returns a pydantic type adapter."""
    from advanced_alchemy.typing import TypeAdapter

    if PYDANTIC_USE_FAILFAST:
        return TypeAdapter(Annotated[f, FailFast()])  # pyright: ignore
    return TypeAdapter(f)


@lru_cache(maxsize=128, typed=True)
def get_attrs_fields(cls: Any) -> "tuple[Any, ...]":
    """Caches and returns attrs fields for a given attrs class."""
    if ATTRS_INSTALLED:
        return fields(cls)  # type: ignore[no-any-return]
    return ()


def is_dto_data(v: Any) -> TypeGuard[DTODataLike[Any]]:
    """Check if a value is a Litestar DTOData object."""
    return LITESTAR_INSTALLED and isinstance(v, DTOData)


def is_pydantic_model(v: Any) -> TypeGuard[BaseModelLike]:
    """Check if a value is a pydantic model."""
    if not PYDANTIC_INSTALLED:
        return False
    if isinstance(v, type):
        try:
            return issubclass(v, BaseModel)
        except TypeError:
            return False
    return isinstance(v, BaseModel)


def is_sqlmodel_table_model(v: Any) -> bool:
    """Check if a value is a SQLModel table model instance or class."""
    if not SQLMODEL_INSTALLED:
        return False
    if isinstance(v, type):
        try:
            return issubclass(v, BaseModel) and hasattr(v, "__mapper__")
        except TypeError:
            return False
    return isinstance(v, BaseModel) and hasattr(v, "__mapper__")


def is_msgspec_struct(v: Any) -> TypeGuard[StructLike]:
    """Check if a value is a msgspec struct."""
    return MSGSPEC_INSTALLED and isinstance(v, Struct)


def is_attrs_instance(obj: Any) -> TypeGuard[AttrsLike]:
    """Check if a value is an attrs class instance."""
    return ATTRS_INSTALLED and has(obj.__class__)


def is_attrs_schema(cls: Any) -> TypeGuard["type[AttrsLike]"]:
    """Check if a class type is an attrs schema."""
    return ATTRS_INSTALLED and has(cls)


def is_dataclass(obj: Any) -> TypeGuard[Any]:
    """Check if an object is a dataclass."""
    return hasattr(obj, "__dataclass_fields__")


def is_dataclass_with_field(obj: Any, field_name: str) -> TypeGuard[object]:
    """Check if an object is a dataclass and has a specific field."""
    return is_dataclass(obj) and hasattr(obj, field_name)


def is_dataclass_without_field(obj: Any, field_name: str) -> TypeGuard[object]:
    """Check if an object is a dataclass and does not have a specific field."""
    return is_dataclass(obj) and not hasattr(obj, field_name)


def is_attrs_instance_with_field(v: Any, field_name: str) -> TypeGuard[AttrsLike]:
    """Check if an attrs instance has a specific field."""
    return is_attrs_instance(v) and hasattr(v, field_name)


def is_attrs_instance_without_field(v: Any, field_name: str) -> TypeGuard[AttrsLike]:
    """Check if an attrs instance does not have a specific field."""
    return is_attrs_instance(v) and not hasattr(v, field_name)


def is_dict(v: Any) -> TypeGuard[dict[str, Any]]:
    """Check if a value is a dictionary."""
    return isinstance(v, dict)


def has_dict_attribute(obj: Any) -> "TypeGuard[_HasDictAttribute]":
    """Check if an object has a __dict__ attribute."""
    return obj is not None and hasattr(obj, "__dict__")


def is_row_mapping(v: Any) -> TypeGuard["RowMapping"]:
    """Check if a value is a SQLAlchemy RowMapping."""
    return isinstance(v, RowMapping)


def is_dict_with_field(v: Any, field_name: str) -> TypeGuard[dict[str, Any]]:
    """Check if a dictionary has a specific field."""
    return is_dict(v) and field_name in v


def is_dict_without_field(v: Any, field_name: str) -> TypeGuard[dict[str, Any]]:
    """Check if a dictionary does not have a specific field."""
    return is_dict(v) and field_name not in v


def is_pydantic_model_with_field(v: Any, field_name: str) -> TypeGuard[BaseModelLike]:
    """Check if a pydantic model has a specific field."""
    return is_pydantic_model(v) and hasattr(v, field_name)


def is_pydantic_model_without_field(v: Any, field_name: str) -> TypeGuard[BaseModelLike]:
    """Check if a pydantic model does not have a specific field."""
    return is_pydantic_model(v) and not hasattr(v, field_name)


def is_msgspec_struct_with_field(v: Any, field_name: str) -> TypeGuard[StructLike]:
    """Check if a msgspec struct has a specific field."""
    return is_msgspec_struct(v) and hasattr(v, field_name)


def is_msgspec_struct_without_field(v: Any, field_name: str) -> "TypeGuard[StructLike]":
    """Check if a msgspec struct does not have a specific field."""
    return is_msgspec_struct(v) and not hasattr(v, field_name)


def is_schema(v: Any) -> "TypeGuard[SupportedSchemaModel]":
    """Check if a value is a msgspec Struct, Pydantic model, or attrs instance."""
    if is_sqlmodel_table_model(v):
        return False
    return is_msgspec_struct(v) or is_pydantic_model(v) or is_attrs_instance(v)


def is_schema_or_dict(v: Any) -> "TypeGuard[Union[SupportedSchemaModel, dict[str, Any]]]":
    """Check if a value is a msgspec Struct, Pydantic model, attrs class, or dict."""
    return is_schema(v) or is_dict(v)


def is_schema_with_field(v: Any, field_name: str) -> "TypeGuard[SupportedSchemaModel]":
    """Check if a value is a msgspec Struct, Pydantic model, or attrs instance with a specific field."""
    if is_sqlmodel_table_model(v):
        return False
    return (
        is_msgspec_struct_with_field(v, field_name)
        or is_pydantic_model_with_field(v, field_name)
        or is_attrs_instance_with_field(v, field_name)
    )


def is_schema_without_field(v: Any, field_name: str) -> "TypeGuard[SupportedSchemaModel]":
    """Check if a value is a msgspec Struct, Pydantic model, or attrs instance without a specific field."""
    return is_schema(v) and not hasattr(v, field_name)


def is_schema_or_dict_with_field(v: Any, field_name: str) -> "TypeGuard[Union[SupportedSchemaModel, dict[str, Any]]]":
    """Check if a value is a msgspec Struct, Pydantic model, attrs instance, or dict with a specific field."""
    return is_schema_with_field(v, field_name) or is_dict_with_field(v, field_name)


def is_schema_or_dict_without_field(
    v: Any, field_name: str
) -> "TypeGuard[Union[SupportedSchemaModel, dict[str, Any]]]":
    """Check if a value is a msgspec Struct, Pydantic model, attrs instance, or dict without a specific field."""
    return is_schema_or_dict(v) and not is_schema_or_dict_with_field(v, field_name)


def _dump_pydantic_model(data: BaseModelLike, dump_config: SchemaDumpConfig) -> "dict[str, Any]":
    """Dump a Pydantic model according to the schema dump config."""
    dumped = data.model_dump(
        exclude_unset=dump_config.exclude_unset,
        exclude_none=dump_config.exclude_none,
        exclude_defaults=dump_config.exclude_defaults,
    )
    if not dump_config.exclude_sentinels and (missing := _get_pydantic_missing_sentinel()) is not None:
        for field_name in getattr(data.__class__, "model_fields", {}):
            if getattr(data, field_name, None) is missing:
                dumped[field_name] = missing
    return dumped


def _dump_msgspec_struct(data: StructLike, dump_config: SchemaDumpConfig) -> "dict[str, Any]":
    """Dump a msgspec struct according to the schema dump config."""
    fields = data.__struct_fields__
    defaults = getattr(data, "__struct_defaults__", ())
    default_fields = fields[-len(defaults) :] if defaults else ()
    default_values = dict(zip(default_fields, defaults))
    dumped: dict[str, Any] = {}
    for field_name in fields:
        if not hasattr(data, field_name):
            continue
        value = getattr(data, field_name)
        if dump_config.exclude_sentinels and value is UNSET:
            continue
        if dump_config.exclude_none and value is None:
            continue
        if dump_config.exclude_defaults and field_name in default_values and value == default_values[field_name]:
            continue
        dumped[field_name] = value
    return dumped


def _dump_attrs_instance(data: AttrsLike, dump_config: SchemaDumpConfig) -> "dict[str, Any]":
    """Dump an attrs instance according to the schema dump config."""
    if dump_config.exclude_sentinels or dump_config.exclude_none:
        # Filter out attrs.NOTHING values for partial updates.
        def filter_unset_attrs(attr: Any, value: Any) -> bool:
            if dump_config.exclude_sentinels and value is attrs_nothing:
                return False
            if dump_config.exclude_none and value is None:
                return False
            return not (dump_config.exclude_defaults and attr.default is not attrs_nothing and value == attr.default)

        return asdict(data, filter=filter_unset_attrs)
    if CATTRS_INSTALLED:
        return unstructure(data)  # type: ignore[no-any-return]
    return asdict(data)


def _dump_dataclass_instance(data: Any, dump_config: SchemaDumpConfig) -> "dict[str, Any]":
    """Dump a dataclass instance according to the schema dump config."""
    dumped: dict[str, Any] = {}
    for field in dataclass_fields(data):
        value = getattr(data, field.name)
        if dump_config.exclude_sentinels and value is DataclassEmpty:
            continue
        if dump_config.exclude_none and value is None:
            continue
        if dump_config.exclude_defaults and field.default is not DATACLASS_MISSING and value == field.default:
            continue
        if is_dataclass_instance(value):
            value = _dump_dataclass_instance(value, dump_config)
        dumped[field.name] = value
    return dumped


def _dump_schema_model(data: Any, dump_config: SchemaDumpConfig) -> "Optional[dict[str, Any]]":
    """Dump a supported schema-like object according to the schema dump config."""
    if is_pydantic_model(data):
        return _dump_pydantic_model(data, dump_config)
    if is_msgspec_struct(data):
        return _dump_msgspec_struct(data, dump_config)
    if is_attrs_instance(data):
        return _dump_attrs_instance(data, dump_config)
    if is_dataclass(data):
        return _dump_dataclass_instance(data, dump_config)
    return None


@overload
def schema_dump(
    data: "RowMapping",
    exclude_unset: bool = True,
    *,
    config: "Optional[SchemaDumpConfig]" = None,
) -> "dict[str, Any]": ...


@overload
def schema_dump(
    data: "Row[Any]",
    exclude_unset: bool = True,
    *,
    config: "Optional[SchemaDumpConfig]" = None,
) -> "dict[str, Any]": ...


@overload
def schema_dump(
    data: "DTODataLike[Any]",
    exclude_unset: bool = True,
    *,
    config: "Optional[SchemaDumpConfig]" = None,
) -> "dict[str, Any]": ...


@overload
def schema_dump(  # pyright: ignore[reportOverlappingOverload]
    data: "ModelT",
    exclude_unset: bool = True,
    *,
    config: "Optional[SchemaDumpConfig]" = None,
) -> "ModelT": ...


@overload
def schema_dump(
    data: Any,
    exclude_unset: bool = True,
    *,
    config: "Optional[SchemaDumpConfig]" = None,
) -> "dict[str, Any]": ...


def schema_dump(  # noqa: PLR0911
    data: "Union[dict[str, Any], ModelT, SupportedSchemaModel, DTODataLike[ModelT], RowMapping, Row[Any]]",
    exclude_unset: bool = True,
    *,
    config: "Optional[SchemaDumpConfig]" = None,
) -> "Union[dict[str, Any], ModelT]":
    """Dump a data object to a dictionary."""
    dump_config = _get_schema_dump_config(exclude_unset=exclude_unset, config=config)
    if is_dict(data):
        return data
    if is_row_mapping(data):
        return dict(data)
    if is_sqlmodel_table_model(data):
        return cast("ModelT", data)
    if (schema_data := _dump_schema_model(data, dump_config)) is not None:
        return schema_data
    if is_dto_data(data):
        return cast("dict[str, Any]", data.as_builtins())
    if has_dict_attribute(data):
        return data.__dict__
    return cast("ModelT", data)


# ============================================================================
# Default type encoders
# ============================================================================


def _build_default_type_encoders() -> dict[type, Callable[[Any], Any]]:
    """Build the default type encoders dictionary.

    This function constructs the default encoders, including optional
    encoders for numpy and pydantic if those packages are installed.

    Returns:
        Dictionary mapping types to encoder functions.
    """
    encoders: dict[type, Callable[[Any], Any]] = {
        # Datetime types
        datetime.datetime: convert_datetime_to_gmt_iso,
        datetime.date: convert_date_to_iso,
        datetime.time: lambda v: v.isoformat(),
        datetime.timedelta: lambda v: v.total_seconds(),
        # Numeric types
        Decimal: str,  # Preserve precision as string
        # UUID and path types
        UUID: str,
        Path: str,
        PurePath: str,
        # Network types
        IPv4Address: str,
        IPv4Interface: str,
        IPv4Network: str,
        IPv6Address: str,
        IPv6Interface: str,
        IPv6Network: str,
        # Collection types
        frozenset: list,
        set: list,
        # Bytes
        bytes: lambda v: v.decode("utf-8", errors="replace"),
        # Enum (use .value to get the underlying value)
        enum.Enum: lambda v: v.value,
    }

    # Optional NumPy support
    if NUMPY_INSTALLED:
        import numpy as np

        encoders[np.ndarray] = lambda v: v.tolist()
        encoders[np.integer] = int
        encoders[np.floating] = float
        encoders[np.bool_] = bool

    # Optional Pydantic support
    if PYDANTIC_INSTALLED:
        from pydantic import BaseModel

        encoders[BaseModel] = lambda v: v.model_dump(mode="json")

    return encoders


DEFAULT_TYPE_ENCODERS: Final[dict[type, Callable[[Any], Any]]] = _build_default_type_encoders()
"""Default type encoders for common Python types.

These encoders handle serialization of types not natively supported
by JSON serializers. Users can override these by passing custom
``type_encoders`` to serializer functions.

Supported types:
    - ``datetime.datetime``: ISO 8601 with UTC timezone (Z suffix)
    - ``datetime.date``: ISO 8601 date format
    - ``datetime.time``: ISO 8601 time format
    - ``datetime.timedelta``: Total seconds as float
    - ``Decimal``: String representation (preserves precision)
    - ``UUID``: String representation
    - ``Path``, ``PurePath``: String representation
    - ``IPv4Address``, ``IPv6Address``, etc.: String representation
    - ``set``, ``frozenset``: Converted to list
    - ``bytes``: UTF-8 decoded string
    - ``enum.Enum``: Value of the enum member
    - ``numpy.ndarray``: Converted to list (if numpy installed)
    - ``pydantic.BaseModel``: model_dump with mode="json" (if pydantic installed)
"""


# ============================================================================
# Protocol and base classes
# ============================================================================


@runtime_checkable
class JSONSerializer(Protocol):
    """Protocol defining the JSON serialization interface.

    Implement this protocol to create custom serializers that integrate
    with Advanced Alchemy's serialization system.

    Example:
        >>> class MySerializer:
        ...     def encode(
        ...         self, data: Any, *, as_bytes: bool = False
        ...     ) -> "Union[str, bytes]":
        ...         # Custom encoding logic
        ...         pass
        ...
        ...     def decode(
        ...         self,
        ...         data: "Union[str, bytes]",
        ...         *,
        ...         decode_bytes: bool = True,
        ...     ) -> Any:
        ...         # Custom decoding logic
        ...         pass
    """

    @overload
    def encode(self, data: Any, *, as_bytes: Literal[False] = ...) -> str: ...

    @overload
    def encode(self, data: Any, *, as_bytes: Literal[True]) -> bytes: ...

    @overload
    def encode(self, data: Any, *, as_bytes: bool) -> "Union[str, bytes]": ...

    def encode(self, data: Any, *, as_bytes: bool = False) -> "Union[str, bytes]":
        """Encode data to JSON.

        Args:
            data: Data to encode.
            as_bytes: If True, return bytes; otherwise return str.

        Returns:
            JSON string or bytes representation.
        """
        ...

    def decode(self, data: "Union[str, bytes]", *, decode_bytes: bool = True) -> Any:
        """Decode JSON to Python object.

        Args:
            data: JSON string or bytes to decode.
            decode_bytes: If True, decode bytes input; otherwise return as-is.

        Returns:
            Decoded Python object.
        """
        ...


class BaseJSONSerializer(ABC):
    """Abstract base class for JSON serializers.

    Provides common functionality for serializer implementations including
    type encoder merging and enc_hook creation.
    """

    __slots__ = ("_custom_type_encoders", "_type_encoders")

    def __init__(self, type_encoders: "Optional[TypeEncodersMap]" = None) -> None:
        """Initialize serializer with optional custom type encoders.

        Args:
            type_encoders: Custom type encoders to merge with defaults.
                User-provided encoders take precedence over defaults.
        """
        self._custom_type_encoders: dict[type, Callable[[Any], Any]] = dict(type_encoders or {})
        self._type_encoders: dict[type, Callable[[Any], Any]] = {
            **DEFAULT_TYPE_ENCODERS,
            **(type_encoders or {}),
        }

    @staticmethod
    def _get_type_encoder(
        value: Any,
        type_encoders: Mapping[type, Callable[[Any], Any]],
    ) -> "Optional[Callable[[Any], Any]]":
        """Return the encoder matching ``value`` by exact type or MRO."""
        for base in value.__class__.__mro__[:-1]:
            encoder = type_encoders.get(base)
            if encoder is not None:
                return encoder
        return None

    def _encode_mapping_key_with_custom_type_encoder(self, key: Any) -> Any:
        encoder = self._get_type_encoder(key, self._custom_type_encoders)
        return encoder(key) if encoder is not None else key

    def _apply_custom_type_encoders_to_mapping(self, value: Mapping[object, object]) -> dict[Any, Any]:
        return {
            self._encode_mapping_key_with_custom_type_encoder(key): self._apply_custom_type_encoders(item)
            for key, item in value.items()
        }

    def _apply_custom_type_encoders_to_list(self, value: list[object]) -> list[Any]:
        return [self._apply_custom_type_encoders(item) for item in value]

    def _apply_custom_type_encoders_to_tuple(self, value: tuple[object, ...]) -> tuple[Any, ...]:
        return tuple(self._apply_custom_type_encoders(item) for item in value)

    def _apply_custom_type_encoders_to_set(self, value: Union[set[object], frozenset[object]]) -> list[Any]:
        return [self._apply_custom_type_encoders(item) for item in value]

    def _apply_custom_type_encoders(self, value: Any) -> Any:
        """Apply user-provided type encoders before backend-native encoding."""
        encoder = self._get_type_encoder(value, self._custom_type_encoders)
        if encoder is not None:
            return encoder(value)

        if isinstance(value, Mapping):
            return self._apply_custom_type_encoders_to_mapping(cast("Mapping[object, object]", value))
        if isinstance(value, list):
            return self._apply_custom_type_encoders_to_list(cast("list[object]", value))
        if isinstance(value, tuple):
            return self._apply_custom_type_encoders_to_tuple(cast("tuple[object, ...]", value))
        if isinstance(value, (set, frozenset)):
            return self._apply_custom_type_encoders_to_set(cast("Union[set[object], frozenset[object]]", value))
        return value

    def _prepare_data_for_encode(self, data: Any) -> Any:
        if not self._custom_type_encoders:
            return data
        return self._apply_custom_type_encoders(data)

    def _create_enc_hook(self) -> "Callable[[Any], Any]":
        """Create an encoding hook function from type_encoders.

        The hook walks the MRO of the value's type to find a matching
        encoder, allowing encoders for parent classes to handle subclasses.

        Returns:
            Encoding hook function suitable for msgspec/orjson.
        """
        type_encoders = self._type_encoders

        def enc_hook(value: Any) -> Any:
            encoder = self._get_type_encoder(value, type_encoders)
            if encoder is not None:
                return encoder(value)
            # Fallback: try string conversion
            try:
                return str(value)
            except Exception as exc:
                msg = f"Cannot serialize {type(value).__name__}"
                raise TypeError(msg) from exc

        return enc_hook

    @overload
    @abstractmethod
    def encode(self, data: Any, *, as_bytes: Literal[False] = ...) -> str: ...

    @overload
    @abstractmethod
    def encode(self, data: Any, *, as_bytes: Literal[True]) -> bytes: ...

    @overload
    @abstractmethod
    def encode(self, data: Any, *, as_bytes: bool) -> "Union[str, bytes]": ...

    @abstractmethod
    def encode(self, data: Any, *, as_bytes: bool = False) -> "Union[str, bytes]":
        """Encode data to JSON."""
        ...

    @abstractmethod
    def decode(self, data: "Union[str, bytes]", *, decode_bytes: bool = True) -> Any:
        """Decode JSON to Python object."""
        ...


# ============================================================================
# Serializer implementations
# ============================================================================


class MsgspecSerializer(BaseJSONSerializer):
    """High-performance JSON serializer using msgspec.

    This is the preferred serializer when msgspec is available, offering
    the best performance for JSON encoding/decoding.

    Example:
        >>> serializer = MsgspecSerializer()
        >>> serializer.encode({"key": "value"})
        '{"key":"value"}'

        >>> # With custom type encoders
        >>> serializer = MsgspecSerializer(type_encoders={set: sorted})
        >>> serializer.encode({"items": {3, 1, 2}})
        '{"items":[1,2,3]}'
    """

    __slots__ = ("_decoder", "_enc_hook", "_encoder")

    def __init__(
        self,
        type_encoders: "Optional[TypeEncodersMap]" = None,
        enc_hook: "Optional[Callable[[Any], Any]]" = None,
    ) -> None:
        """Initialize msgspec serializer.

        Args:
            type_encoders: Custom type encoders to merge with defaults.
            enc_hook: Optional custom encoding hook. If provided, type_encoders
                is ignored and this hook is used directly for complete control.
        """
        from msgspec.json import Decoder, Encoder

        super().__init__(None if enc_hook is not None else type_encoders)

        self._enc_hook = enc_hook if enc_hook is not None else self._create_enc_hook()
        self._encoder: Final = Encoder(enc_hook=self._enc_hook)
        self._decoder: Final = Decoder()

    @overload
    def encode(self, data: Any, *, as_bytes: Literal[False] = ...) -> str: ...

    @overload
    def encode(self, data: Any, *, as_bytes: Literal[True]) -> bytes: ...

    @overload
    def encode(self, data: Any, *, as_bytes: bool) -> "Union[str, bytes]": ...

    def encode(self, data: Any, *, as_bytes: bool = False) -> "Union[str, bytes]":
        """Encode data to JSON using msgspec.

        Args:
            data: Data to encode.
            as_bytes: If True, return bytes; otherwise return UTF-8 string.

        Returns:
            JSON representation as string or bytes.
        """
        result = self._encoder.encode(self._prepare_data_for_encode(data))
        return result if as_bytes else result.decode("utf-8")

    def decode(self, data: "Union[str, bytes]", *, decode_bytes: bool = True) -> Any:
        """Decode JSON using msgspec.

        Args:
            data: JSON string or bytes to decode.
            decode_bytes: If True, decode bytes input; otherwise return as-is.

        Returns:
            Decoded Python object.
        """
        if isinstance(data, str):
            data = data.encode("utf-8")
        elif not decode_bytes:
            return data
        return self._decoder.decode(data)


class OrjsonSerializer(BaseJSONSerializer):
    """High-performance JSON serializer using orjson.

    Provides excellent performance with native support for datetime,
    UUID, and numpy serialization.

    Example:
        >>> serializer = OrjsonSerializer()
        >>> serializer.encode({"key": "value"})
        '{"key":"value"}'
    """

    __slots__ = ("_enc_hook",)

    def __init__(self, type_encoders: "Optional[TypeEncodersMap]" = None) -> None:
        """Initialize orjson serializer.

        Args:
            type_encoders: Custom type encoders to merge with defaults.
        """
        super().__init__(type_encoders)
        self._enc_hook = self._create_enc_hook()

    @overload
    def encode(self, data: Any, *, as_bytes: Literal[False] = ...) -> str: ...

    @overload
    def encode(self, data: Any, *, as_bytes: Literal[True]) -> bytes: ...

    @overload
    def encode(self, data: Any, *, as_bytes: bool) -> "Union[str, bytes]": ...

    def encode(self, data: Any, *, as_bytes: bool = False) -> "Union[str, bytes]":
        """Encode data to JSON using orjson.

        Args:
            data: Data to encode.
            as_bytes: If True, return bytes; otherwise return UTF-8 string.

        Returns:
            JSON representation as string or bytes.
        """
        import orjson  # type: ignore[import-not-found,unused-ignore]  # pyright: ignore[reportMissingImports]

        orjson_module: Any = orjson
        options: int = orjson_module.OPT_NAIVE_UTC | orjson_module.OPT_SERIALIZE_UUID
        if NUMPY_INSTALLED:
            options |= orjson_module.OPT_SERIALIZE_NUMPY

        result: bytes = orjson_module.dumps(self._prepare_data_for_encode(data), default=self._enc_hook, option=options)
        return result if as_bytes else result.decode("utf-8")

    def decode(self, data: "Union[str, bytes]", *, decode_bytes: bool = True) -> Any:
        """Decode JSON using orjson.

        Args:
            data: JSON string or bytes to decode.
            decode_bytes: If True, decode bytes input; otherwise return as-is.

        Returns:
            Decoded Python object.
        """
        import orjson  # type: ignore[import-not-found,unused-ignore]  # pyright: ignore[reportMissingImports]

        orjson_module: Any = orjson
        if isinstance(data, bytes) and not decode_bytes:
            return data
        return orjson_module.loads(data)


class StandardLibSerializer(BaseJSONSerializer):
    """JSON serializer using Python's standard library.

    Fallback serializer when neither msgspec nor orjson is available.
    Slower than the alternatives but always available.

    Example:
        >>> serializer = StandardLibSerializer()
        >>> serializer.encode({"key": "value"})
        '{"key": "value"}'
    """

    __slots__ = ("_enc_hook",)

    def __init__(self, type_encoders: "Optional[TypeEncodersMap]" = None) -> None:
        """Initialize standard library serializer.

        Args:
            type_encoders: Custom type encoders to merge with defaults.
        """
        super().__init__(type_encoders)
        self._enc_hook = self._create_enc_hook()

    @overload
    def encode(self, data: Any, *, as_bytes: Literal[False] = ...) -> str: ...

    @overload
    def encode(self, data: Any, *, as_bytes: Literal[True]) -> bytes: ...

    @overload
    def encode(self, data: Any, *, as_bytes: bool) -> "Union[str, bytes]": ...

    def encode(self, data: Any, *, as_bytes: bool = False) -> "Union[str, bytes]":
        """Encode data to JSON using standard library.

        Args:
            data: Data to encode.
            as_bytes: If True, return bytes; otherwise return string.

        Returns:
            JSON representation as string or bytes.
        """
        result = json.dumps(self._prepare_data_for_encode(data), default=self._enc_hook)
        return result.encode("utf-8") if as_bytes else result

    def decode(self, data: "Union[str, bytes]", *, decode_bytes: bool = True) -> Any:
        """Decode JSON using standard library.

        Args:
            data: JSON string or bytes to decode.
            decode_bytes: If True, decode bytes input; otherwise return as-is.

        Returns:
            Decoded Python object.
        """
        if isinstance(data, bytes):
            if not decode_bytes:
                return data
            data = data.decode("utf-8")
        return json.loads(data)


# ============================================================================
# Serializer factory with caching
# ============================================================================

_serializer_cache: "dict[frozenset[tuple[type, int]], JSONSerializer]" = {}
_cache_lock: Final[threading.RLock] = threading.RLock()
_default_serializer: "Optional[JSONSerializer]" = None


def _create_default_serializer() -> "JSONSerializer":
    """Create the default serializer based on available libraries.

    Priority: msgspec > orjson > standard library

    Returns:
        Best available JSON serializer.
    """
    if MSGSPEC_INSTALLED:
        return MsgspecSerializer()
    if ORJSON_INSTALLED:
        return OrjsonSerializer()
    return StandardLibSerializer()


def _create_serializer(type_encoders: "TypeEncodersMap") -> "JSONSerializer":
    """Create a serializer with custom type encoders.

    Args:
        type_encoders: Custom type encoders mapping.

    Returns:
        JSON serializer configured with the given encoders.
    """
    if MSGSPEC_INSTALLED:
        return MsgspecSerializer(type_encoders=type_encoders)
    if ORJSON_INSTALLED:
        return OrjsonSerializer(type_encoders=type_encoders)
    return StandardLibSerializer(type_encoders=type_encoders)


def get_serializer(type_encoders: "Optional[TypeEncodersMap]" = None) -> "JSONSerializer":
    """Get a cached serializer instance.

    This factory function returns cached serializer instances to avoid
    recreating encoders for the same configuration. The default serializer
    (no custom type_encoders) is a singleton for maximum performance.

    Args:
        type_encoders: Optional mapping of types to encoder functions.
            If None, returns the default singleton serializer.

    Returns:
        A JSONSerializer instance configured with the given type encoders.

    Example:
        >>> # Default serializer (singleton)
        >>> s1 = get_serializer()
        >>> s2 = get_serializer()
        >>> s1 is s2
        True

        >>> # Custom type encoders (cached by configuration)
        >>> encoders = {str: lambda s: s.upper()}
        >>> s3 = get_serializer(encoders)
        >>> s4 = get_serializer(encoders)
        >>> s3 is s4
        True
    """
    global _default_serializer  # noqa: PLW0603

    if type_encoders is None:
        with _cache_lock:
            if _default_serializer is None:
                _default_serializer = _create_default_serializer()
            return _default_serializer

    # Create hashable cache key using type and function id
    # Note: Different lambda objects create different keys even if equivalent
    try:
        cache_key = frozenset((k, id(v)) for k, v in type_encoders.items())
    except TypeError:
        # Fallback for unhashable keys - create without caching
        return _create_serializer(type_encoders)

    with _cache_lock:
        if cache_key not in _serializer_cache:
            _serializer_cache[cache_key] = _create_serializer(type_encoders)
        return _serializer_cache[cache_key]


# ============================================================================
# Public API functions
# ============================================================================


@overload
def encode_json(data: Any) -> str: ...


@overload
def encode_json(data: Any, *, as_bytes: Literal[False]) -> str: ...


@overload
def encode_json(data: Any, *, as_bytes: Literal[True]) -> bytes: ...


@overload
def encode_json(data: Any, *, type_encoders: "TypeEncodersMap") -> str: ...


@overload
def encode_json(data: Any, *, type_encoders: TypeEncodersMap, as_bytes: Literal[False]) -> str: ...


@overload
def encode_json(data: Any, *, type_encoders: TypeEncodersMap, as_bytes: Literal[True]) -> bytes: ...


def encode_json(
    data: Any,
    *,
    type_encoders: "Optional[TypeEncodersMap]" = None,
    as_bytes: bool = False,
) -> "Union[str, bytes]":
    """Encode data to JSON with optional custom type encoders.

    This is the primary interface for JSON encoding in Advanced Alchemy.
    It supports custom type encoders that can be passed per-call.

    Args:
        data: Data to encode to JSON.
        type_encoders: Optional mapping of types to encoder functions.
            These are merged with DEFAULT_TYPE_ENCODERS, with user
            encoders taking precedence.
        as_bytes: If True, return bytes; otherwise return str.

    Returns:
        JSON representation of data as string or bytes.

    Example:
        >>> import datetime
        >>> encode_json({"date": datetime.date(2025, 12, 17)})
        '{"date":"2025-12-17"}'

        >>> # Custom type encoder
        >>> class Point:
        ...     def __init__(self, x, y):
        ...         self.x, self.y = x, y
        >>> encode_json(
        ...     {"point": Point(1, 2)},
        ...     type_encoders={Point: lambda p: [p.x, p.y]},
        ... )
        '{"point":[1,2]}'

        >>> # Return bytes
        >>> encode_json({"key": "value"}, as_bytes=True)
        b'{"key":"value"}'
    """
    serializer = get_serializer(type_encoders)
    return serializer.encode(data, as_bytes=as_bytes)


def decode_json(data: "Union[str, bytes]", *, decode_bytes: bool = True) -> Any:
    """Decode JSON string or bytes to Python object.

    Args:
        data: JSON string or bytes to decode.
        decode_bytes: If True, decode bytes input. If False, bytes
            are returned as-is without decoding.

    Returns:
        Decoded Python object.

    Example:
        >>> decode_json('{"key": "value"}')
        {'key': 'value'}

        >>> decode_json(b'{"key": "value"}')
        {'key': 'value'}
    """
    serializer = get_serializer()
    return serializer.decode(data, decode_bytes=decode_bytes)


def encode_complex_type(obj: Any) -> Any:  # noqa: PLR0911
    """Convert an object to a JSON-serializable format if possible.

    Handles types that are not natively JSON serializable:
    - datetime, date, time: ISO format strings
    - timedelta: total seconds as float
    - Decimal: string representation
    - bytes: hex string
    - UUID: string representation
    - set, frozenset: list

    Args:
        obj: The object to encode.

    Returns:
        A JSON-serializable representation of the object, or None if the type is not supported.
    """
    if isinstance(obj, datetime.datetime):
        return {"__type__": "datetime", "value": obj.isoformat()}
    if isinstance(obj, datetime.date):
        return {"__type__": "date", "value": obj.isoformat()}
    if isinstance(obj, datetime.time):
        return {"__type__": "time", "value": obj.isoformat()}
    if isinstance(obj, datetime.timedelta):
        return {"__type__": "timedelta", "value": obj.total_seconds()}
    if isinstance(obj, Decimal):
        return {"__type__": "decimal", "value": str(obj)}
    if isinstance(obj, bytes):
        return {"__type__": "bytes", "value": obj.hex()}
    if isinstance(obj, UUID):
        return {"__type__": "uuid", "value": str(obj)}
    if isinstance(obj, (set, frozenset)):
        items: list[Any] = list(cast("Union[set[Any], frozenset[Any]]", obj))  # type: ignore[redundant-cast]
        return {"__type__": "set", "value": items}
    return None


def decode_complex_type(value: Any) -> Any:
    """Recursively decode special type markers.

    Decodes the special `{"__type__": ..., "value": ...}` structures.
    """
    if isinstance(value, list):
        value_list = cast("list[Any]", value)  # type: ignore[redundant-cast]
        return [decode_complex_type(v) for v in value_list]

    if not isinstance(value, dict):
        return value

    # Decode any nested values first
    value_dict = cast("dict[Any, Any]", value)  # type: ignore[redundant-cast]
    decoded: dict[str, Any] = {str(k): decode_complex_type(v) for k, v in value_dict.items()}

    # Then decode "typed" marker dicts
    if "__type__" in decoded and "value" in decoded:
        return _decode_typed_marker(decoded)

    return decoded


def _decode_typed_marker(obj: dict[str, Any]) -> Any:  # noqa: PLR0911
    """Custom JSON decoder for special types.

    Args:
        obj: The dictionary to decode.

    Returns:
        The decoded object, or the original dict if not a special type.
    """
    type_name = obj["__type__"]
    value = obj["value"]

    if type_name == "datetime":
        return datetime.datetime.fromisoformat(value)
    if type_name == "date":
        return datetime.date.fromisoformat(value)
    if type_name == "time":
        return datetime.time.fromisoformat(value)
    if type_name == "timedelta":
        return datetime.timedelta(seconds=value)
    if type_name == "decimal":
        return Decimal(value)
    if type_name == "bytes":
        return bytes.fromhex(value)
    if type_name == "uuid":
        return UUID(value)
    if type_name == "set":
        return set(value)

    return obj



import datetime
from collections.abc import Awaitable, Generator
from typing import Any, Callable, TypeVar
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from litestar.exceptions import ImproperlyConfiguredException
from litestar.types import Empty
from pytest import MonkeyPatch
from sqlalchemy import Dialect
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session as SyncSession

from advanced_alchemy.extensions.litestar.plugins.init.config.asyncio import SQLAlchemyAsyncConfig
from advanced_alchemy.extensions.litestar.plugins.init.config.sync import SQLAlchemySyncConfig
from advanced_alchemy.extensions.litestar.store import SQLAlchemyStore, StoreModelMixin

# Type variable for the callable in the patch
F = TypeVar("F", bound=Callable[..., Any])


class MockDateTime:
    """Mock datetime class for testing time-dependent behavior."""

    def __init__(self, fixed_time: datetime.datetime) -> None:
        self.fixed_time = fixed_time

    def now(self, tz: datetime.timezone = datetime.timezone.utc) -> datetime.datetime:
        return self.fixed_time


class MockStoreModel(StoreModelMixin):
    """Mock store model for testing."""

    __tablename__ = "mock_store"


@pytest.fixture()
def mock_store_model() -> type[StoreModelMixin]:
    return MockStoreModel


@pytest.fixture()
def mock_async_session() -> AsyncMock:
    session = AsyncMock(spec=AsyncSession)
    session.__aenter__.return_value = session  # Simulate async context manager
    session.__aexit__.return_value = None

    # Set up the dialect for merge/upsert support
    dialect = MagicMock(spec=Dialect)
    dialect.name = "postgresql"
    dialect.server_version_info = (15, 0)
    session.bind = MagicMock()
    session.bind.dialect = dialect

    # Configure execute to return a mock that can be used for scalar_one_or_none
    execute_result = AsyncMock()
    execute_result.scalar_one_or_none = AsyncMock()  # Create an async mock for the method
    session.execute = AsyncMock(return_value=execute_result)  # Create an async mock for execute

    return session


@pytest.fixture()
def mock_async_config(mock_store_model: type[StoreModelMixin], mock_async_session: AsyncMock) -> MagicMock:
    config = MagicMock(spec=SQLAlchemyAsyncConfig)
    config.get_session.return_value = mock_async_session  # Configure the mock method
    return config


@pytest.fixture()
def mock_sync_session() -> MagicMock:
    session = MagicMock(spec=SyncSession)
    session.__enter__.return_value = session  # Simulate sync context manager
    session.__exit__.return_value = None

    # Set up the dialect for merge/upsert support
    dialect = MagicMock(spec=Dialect)
    dialect.name = "postgresql"
    dialect.server_version_info = (15, 0)
    session.bind = MagicMock()
    session.bind.dialect = dialect

    # Configure execute to return a mock that can be used for scalar_one_or_none
    execute_result = MagicMock()
    execute_result.scalar_one_or_none.return_value = None
    session.execute.return_value = execute_result

    return session


@pytest.fixture()
def mock_sync_config(mock_store_model: type[StoreModelMixin], mock_sync_session: MagicMock) -> MagicMock:
    config = MagicMock(spec=SQLAlchemySyncConfig)
    config.get_session.return_value = mock_sync_session  # Configure the mock method
    return config


@pytest.fixture()
def async_store(mock_async_config: MagicMock) -> SQLAlchemyStore[SQLAlchemyAsyncConfig]:
    """Create an async store instance."""
    return SQLAlchemyStore(config=mock_async_config, model=MockStoreModel)


@pytest.fixture()
def sync_store(mock_sync_config: MagicMock) -> SQLAlchemyStore[SQLAlchemySyncConfig]:
    """Create a sync store instance."""
    return SQLAlchemyStore(config=mock_sync_config, model=MockStoreModel)


def test_store_model_mixin_is_expired_property() -> None:
    """Test the is_expired hybrid property."""
    now = datetime.datetime.now(datetime.timezone.utc)
    expired_store = MockStoreModel(expires_at=now - datetime.timedelta(seconds=1))
    active_store = MockStoreModel(expires_at=now + datetime.timedelta(seconds=10))

    assert expired_store.is_expired is True
    assert active_store.is_expired is False


def test_store_init_default_namespace() -> None:
    """Test store initialization with default namespace."""
    store = SQLAlchemyStore(config=MagicMock(spec=SQLAlchemyAsyncConfig), model=MockStoreModel)
    assert store.namespace == "LITESTAR"


def test_store_init_custom_namespace() -> None:
    """Test store initialization with custom namespace."""
    store = SQLAlchemyStore(config=MagicMock(spec=SQLAlchemyAsyncConfig), model=MockStoreModel, namespace="CUSTOM")
    assert store.namespace == "CUSTOM"


def test_store_init_no_namespace() -> None:
    """Test store initialization with no namespace."""
    store = SQLAlchemyStore(config=MagicMock(spec=SQLAlchemyAsyncConfig), model=MockStoreModel, namespace=None)
    assert store.namespace is None


def test_store_init_empty_namespace() -> None:
    """Test store initialization with Empty namespace."""
    store = SQLAlchemyStore(config=MagicMock(spec=SQLAlchemyAsyncConfig), model=MockStoreModel, namespace=Empty)
    assert store.namespace == "LITESTAR"


def test_supports_merge() -> None:
    """Test merge support detection for different dialects."""
    store = SQLAlchemyStore(config=MagicMock(spec=SQLAlchemyAsyncConfig), model=MockStoreModel)

    # PostgreSQL >= 15 (currently disabled via _DISABLE_POSTGRES_MERGE)
    postgres_dialect = MagicMock(spec=Dialect)
    postgres_dialect.name = "postgresql"
    postgres_dialect.server_version_info = (15, 0)
    assert store.supports_merge(postgres_dialect) is False  # Disabled due to _DISABLE_POSTGRES_MERGE = True

    # PostgreSQL < 15
    postgres_dialect.server_version_info = (14, 0)
    assert store.supports_merge(postgres_dialect) is False

    # Oracle
    oracle_dialect = MagicMock(spec=Dialect)
    oracle_dialect.name = "oracle"
    oracle_dialect.server_version_info = (19, 0)  # Add server_version_info for Oracle
    assert store.supports_merge(oracle_dialect) is True

    # Other dialects
    other_dialect = MagicMock(spec=Dialect)
    other_dialect.name = "mysql"
    other_dialect.server_version_info = (8, 0)  # Add server_version_info for MySQL
    assert store.supports_merge(other_dialect) is False


def test_supports_upsert() -> None:
    """Test upsert support detection for different dialects."""
    store = SQLAlchemyStore(config=MagicMock(spec=SQLAlchemyAsyncConfig), model=MockStoreModel)

    # PostgreSQL
    postgres_dialect = MagicMock(spec=Dialect)
    postgres_dialect.name = "postgresql"
    assert store.supports_upsert(postgres_dialect) is True

    # CockroachDB
    cockroach_dialect = MagicMock(spec=Dialect)
    cockroach_dialect.name = "cockroachdb"
    assert store.supports_upsert(cockroach_dialect) is True

    # SQLite
    sqlite_dialect = MagicMock(spec=Dialect)
    sqlite_dialect.name = "sqlite"
    assert store.supports_upsert(sqlite_dialect) is True

    # MySQL
    mysql_dialect = MagicMock(spec=Dialect)
    mysql_dialect.name = "mysql"
    assert store.supports_upsert(mysql_dialect) is True

    # Other dialects
    other_dialect = MagicMock(spec=Dialect)
    other_dialect.name = "other"
    assert store.supports_upsert(other_dialect) is False


def test_make_key_with_namespace() -> None:
    """Test key generation with namespace."""
    store = SQLAlchemyStore(config=MagicMock(spec=SQLAlchemyAsyncConfig), model=MockStoreModel, namespace="test")

    # Instead of using make_key, we'll test the key format directly
    # by checking the namespace and key format
    assert store.namespace == "test"


def test_make_key_without_namespace() -> None:
    """Test key generation without namespace."""
    store = SQLAlchemyStore(config=MagicMock(spec=SQLAlchemyAsyncConfig), model=MockStoreModel, namespace=None)

    # Instead of using make_key, we'll test the key format directly
    # by checking the namespace and key format
    assert store.namespace is None


@pytest.fixture()
def sync_store_with_mock_async(
    mock_sync_config: MagicMock,
) -> Generator[SQLAlchemyStore[SQLAlchemySyncConfig], None, None]:
    def mock_async_(fn: Callable[..., Any], **_: Any) -> Callable[..., Awaitable[Any]]:
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            return fn(*args, **kwargs)

        return wrapper

    with patch("advanced_alchemy.extensions.litestar.store.async_", side_effect=mock_async_):
        yield SQLAlchemyStore(config=mock_sync_config, model=MockStoreModel)


@pytest.mark.asyncio()
async def test_store_set_async(
    async_store: SQLAlchemyStore[SQLAlchemyAsyncConfig], mock_async_session: AsyncMock
) -> None:
    """Test setting a value using async store."""
    key = "test_key"
    value = b"test_value"
    expires_in = 3600

    await async_store.set(key, value, expires_in)

    mock_async_session.execute.assert_called_once()
    await mock_async_session.commit()


@pytest.mark.asyncio()
async def test_store_set_sync(
    sync_store_with_mock_async: SQLAlchemyStore[SQLAlchemySyncConfig], mock_sync_session: MagicMock
) -> None:
    """Test setting a value using sync store."""
    key = "test_key"
    value = b"test_value"
    expires_in = 3600

    # Reset the mock before the test
    mock_sync_session.execute.reset_mock()
    mock_sync_session.commit.reset_mock()

    await sync_store_with_mock_async.set(key, value, expires_in)

    # Verify that execute was called at least once
    assert mock_sync_session.execute.call_count > 0
    assert mock_sync_session.commit.call_count > 0


@pytest.mark.asyncio()
async def test_store_get_async(
    async_store: SQLAlchemyStore[SQLAlchemyAsyncConfig], mock_async_session: AsyncMock
) -> None:
    """Test getting a value using async store."""
    key = "test_key"
    expected_value = b"test_value"

    # Set up the mock to return the expected value directly
    mock_result = MagicMock()
    mock_result.scalar_one_or_none.return_value = expected_value
    mock_async_session.execute.return_value = mock_result

    # Use the async context manager to get the session
    async with async_store:
        result = await async_store.get(key)
        assert result == expected_value
        assert mock_async_session.execute.called
        await mock_async_session.commit()


@pytest.mark.asyncio()
async def test_store_get_sync(
    sync_store_with_mock_async: SQLAlchemyStore[SQLAlchemySyncConfig], mock_sync_session: MagicMock
) -> None:
    """Test getting a value using sync store."""
    key = "test_key"
    expected_value = b"test_value"
    mock_sync_session.execute.return_value.scalar_one_or_none.return_value = expected_value

    result = await sync_store_with_mock_async.get(key)

    assert result == expected_value
    mock_sync_session.execute.assert_called_once()
    mock_sync_session.commit.assert_called_once()


@pytest.mark.asyncio()
async def test_store_delete_async(
    async_store: SQLAlchemyStore[SQLAlchemyAsyncConfig], mock_async_session: AsyncMock
) -> None:
    """Test deleting a value using async store."""
    key = "test_key"

    await async_store.delete(key)

    mock_async_session.execute.assert_called_once()
    await mock_async_session.commit()


@pytest.mark.asyncio()
async def test_store_delete_sync(
    sync_store_with_mock_async: SQLAlchemyStore[SQLAlchemySyncConfig], mock_sync_session: MagicMock
) -> None:
    """Test deleting a value using sync store."""
    key = "test_key"

    await sync_store_with_mock_async.delete(key)

    mock_sync_session.execute.assert_called_once()
    mock_sync_session.commit.assert_called_once()


@pytest.mark.asyncio()
async def test_store_delete_all_async(
    async_store: SQLAlchemyStore[SQLAlchemyAsyncConfig], mock_async_session: AsyncMock
) -> None:
    """Test deleting all values using async store."""
    await async_store.delete_all()

    mock_async_session.execute.assert_called_once()
    await mock_async_session.commit()


@pytest.mark.asyncio()
async def test_store_delete_all_sync(
    sync_store_with_mock_async: SQLAlchemyStore[SQLAlchemySyncConfig], mock_sync_session: MagicMock
) -> None:
    """Test deleting all values using sync store."""
    await sync_store_with_mock_async.delete_all()

    mock_sync_session.execute.assert_called_once()
    mock_sync_session.commit.assert_called_once()


@pytest.mark.asyncio()
async def test_store_delete_all_no_namespace_async(mock_async_config: MagicMock) -> None:
    """Test deleting all values with no namespace raises error."""
    store = SQLAlchemyStore(config=mock_async_config, model=MockStoreModel, namespace=None)
    with pytest.raises(ImproperlyConfiguredException, match="Cannot perform delete operation: No namespace configured"):
        await store.delete_all()


@pytest.mark.asyncio()
async def test_store_delete_all_no_namespace_sync(mock_sync_config: MagicMock) -> None:
    """Test deleting all values with no namespace raises error."""
    store = SQLAlchemyStore(config=mock_sync_config, model=MockStoreModel, namespace=None)
    with pytest.raises(ImproperlyConfiguredException, match="Cannot perform delete operation: No namespace configured"):
        await store.delete_all()


@pytest.mark.asyncio()
async def test_store_exists_async(
    async_store: SQLAlchemyStore[SQLAlchemyAsyncConfig], mock_async_session: AsyncMock
) -> None:
    """Test checking existence using async store."""
    key = "test_key"

    # Set up the mock to return a value directly
    mock_result = MagicMock()
    mock_result.scalar_one_or_none.return_value = "exists"
    mock_async_session.execute.return_value = mock_result

    # Use the async context manager to get the session
    async with async_store:
        result = await async_store.exists(key)
        assert result is True
        assert mock_async_session.execute.called


@pytest.mark.asyncio()
async def test_store_exists_sync(
    sync_store_with_mock_async: SQLAlchemyStore[SQLAlchemySyncConfig], mock_sync_session: MagicMock
) -> None:
    """Test checking existence using sync store."""
    key = "test_key"
    mock_sync_session.execute.return_value.scalar_one_or_none.return_value = "exists"

    result = await sync_store_with_mock_async.exists(key)

    assert result is True
    mock_sync_session.execute.assert_called_once()


@pytest.mark.asyncio()
async def test_store_expires_in_async(
    async_store: SQLAlchemyStore[SQLAlchemyAsyncConfig], mock_async_session: AsyncMock, monkeypatch: MonkeyPatch
) -> None:
    """Test getting expiration time using async store."""
    now = datetime.datetime.now(datetime.timezone.utc)
    expires_at = now + datetime.timedelta(seconds=3600)
    monkeypatch.setattr("advanced_alchemy.extensions.litestar.store.datetime.datetime", MockDateTime(now))

    # Set up the mock to return the expiration time directly
    mock_result = MagicMock()
    mock_result.scalar_one_or_none.return_value = expires_at
    mock_async_session.execute.return_value = mock_result

    # Use the async context manager to get the session
    async with async_store:
        result = await async_store.expires_in("test_key")
        assert result == 3600
        assert mock_async_session.execute.called


@pytest.mark.asyncio()
async def test_store_expires_in_sync(
    sync_store_with_mock_async: SQLAlchemyStore[SQLAlchemySyncConfig],
    mock_sync_session: MagicMock,
    monkeypatch: MonkeyPatch,
) -> None:
    """Test getting expiration time using sync store."""
    key = "test_key"
    now = datetime.datetime.now(datetime.timezone.utc)
    expires_at = now + datetime.timedelta(seconds=3600)
    mock_datetime = MockDateTime(now)
    monkeypatch.setattr("advanced_alchemy.extensions.litestar.store.datetime.datetime", mock_datetime)
    mock_sync_session.execute.return_value.scalar_one_or_none.return_value = expires_at

    result = await sync_store_with_mock_async.expires_in(key)

    assert result == 3600
    mock_sync_session.execute.assert_called_once()


def test_store_with_namespace() -> None:
    """Test creating a new store with nested namespace."""
    store = SQLAlchemyStore(config=MagicMock(spec=SQLAlchemyAsyncConfig), model=MockStoreModel, namespace="base")
    nested_store = store.with_namespace("nested")
    assert nested_store.namespace == "base_nested"

    # Test with no base namespace
    store = SQLAlchemyStore(config=MagicMock(spec=SQLAlchemyAsyncConfig), model=MockStoreModel, namespace=None)
    nested_store = store.with_namespace("nested")
    assert nested_store.namespace == "nested"


@pytest.mark.asyncio()
async def test_store_context_manager() -> None:
    """Test store as async context manager."""
    store = SQLAlchemyStore(config=MagicMock(spec=SQLAlchemyAsyncConfig), model=MockStoreModel)
    async with store:
        pass  # Context manager should not raise any errors



"""Unit tests for the TOTPSecret column type and TOTPProvider wrapper."""

import pytest

from advanced_alchemy.typing import PYOTP_INSTALLED


def test_pyotp_installed_flag_is_bool() -> None:
    assert isinstance(PYOTP_INSTALLED, bool)


@pytest.mark.skipif(not PYOTP_INSTALLED, reason="pyotp not installed")
def test_generate_and_roundtrip_via_provider() -> None:
    import pyotp

    from advanced_alchemy.types import TOTPProvider, generate_totp_secret

    secret = generate_totp_secret()
    provider = TOTPProvider(secret, digits=6, interval=30, digest=None, issuer="ACME")
    code = pyotp.TOTP(secret).now()
    assert provider.verify(code) is True
    assert provider.verify("000000", valid_window=0) in (True, False)
    assert provider.secret == secret
    uri = provider.provisioning_uri(name="alice@example.com")
    assert uri.startswith("otpauth://totp/")
    assert "issuer=ACME" in uri


def test_generate_totp_secret_requires_pyotp(monkeypatch: pytest.MonkeyPatch) -> None:
    from advanced_alchemy.exceptions import MissingDependencyError
    from advanced_alchemy.types import totp

    monkeypatch.setattr(totp, "PYOTP_INSTALLED", False)
    with pytest.raises(MissingDependencyError):
        totp.generate_totp_secret()


def test_totp_secret_requires_pyotp(monkeypatch: pytest.MonkeyPatch) -> None:
    from advanced_alchemy.exceptions import MissingDependencyError
    from advanced_alchemy.types import totp

    monkeypatch.setattr(totp, "PYOTP_INSTALLED", False)
    with pytest.raises(MissingDependencyError):
        totp.TOTPSecret(key="k")


def test_totp_provider_requires_pyotp(monkeypatch: pytest.MonkeyPatch) -> None:
    from advanced_alchemy.exceptions import MissingDependencyError
    from advanced_alchemy.types import totp

    monkeypatch.setattr(totp, "PYOTP_INSTALLED", False)
    with pytest.raises(MissingDependencyError):
        totp.TOTPProvider("ABCDEFGH")


@pytest.mark.skipif(not PYOTP_INSTALLED, reason="pyotp not installed")
def test_totp_secret_requires_explicit_key() -> None:
    """The new type has no deprecated random default; key is required."""
    from advanced_alchemy.types import TOTPSecret

    with pytest.raises(TypeError):
        TOTPSecret()  # type: ignore[call-arg]


@pytest.mark.skipif(not PYOTP_INSTALLED, reason="pyotp not installed")
def test_totp_secret_repr_includes_params() -> None:
    """The repr must name TOTPSecret and carry its TOTP params so Alembic reconstructs faithfully."""
    from advanced_alchemy.types import TOTPSecret

    rendered = repr(TOTPSecret(key="k", issuer="ACME", digits=8, interval=60))
    assert rendered == (
        "TOTPSecret(key='k', backend=FernetBackend, digits=8, interval=60, digest=None, issuer='ACME', length=None)"
    )


@pytest.mark.skipif(not PYOTP_INSTALLED, reason="pyotp not installed")
def test_totp_secret_distinct_params_distinct_cache_key() -> None:
    from advanced_alchemy.types import TOTPSecret

    a = TOTPSecret(key="k", digits=6)
    b = TOTPSecret(key="k", digits=8)
    assert a._static_cache_key != b._static_cache_key



"""Unit tests for the configurable serialization module."""

from __future__ import annotations

import datetime
import importlib
import json
import threading
import time
import warnings
from decimal import Decimal
from enum import Enum
from ipaddress import IPv4Address, IPv4Interface, IPv4Network, IPv6Address, IPv6Interface, IPv6Network
from pathlib import Path, PurePath
from typing import Any
from uuid import UUID

import pytest

from advanced_alchemy.typing import (
    ATTRS_INSTALLED,
    MSGSPEC_INSTALLED,
    NUMPY_INSTALLED,
    ORJSON_INSTALLED,
    PYDANTIC_INSTALLED,
)
from advanced_alchemy.utils import serialization
from advanced_alchemy.utils.dataclass import Empty
from advanced_alchemy.utils.serialization import (
    DEFAULT_TYPE_ENCODERS,
    JSONSerializer,
    MsgspecSerializer,
    OrjsonSerializer,
    SchemaDumpConfig,
    StandardLibSerializer,
    TypeEncodersMap,
    convert_date_to_iso,
    convert_datetime_to_gmt_iso,
    decode_complex_type,
    decode_json,
    encode_complex_type,
    encode_json,
    get_serializer,
    schema_dump,
)


class MyEnum(Enum):
    """Test enum."""

    VALUE_A = "a"
    VALUE_B = "b"


class CustomType:
    """Custom type for testing type encoders."""

    def __init__(self, value: str) -> None:
        self.value = value


class ChildCustomType(CustomType):
    """Child class to test MRO lookup."""


# ----------------------------------------------------------------------
# Helper functions
# ----------------------------------------------------------------------


def test_convert_datetime_to_gmt_iso_naive() -> None:
    dt = datetime.datetime(2025, 12, 17, 10, 30, 0)
    assert convert_datetime_to_gmt_iso(dt) == "2025-12-17T10:30:00Z"


def test_convert_datetime_to_gmt_iso_aware() -> None:
    dt = datetime.datetime(2025, 12, 17, 10, 30, 0, tzinfo=datetime.timezone.utc)
    assert convert_datetime_to_gmt_iso(dt) == "2025-12-17T10:30:00Z"


def test_convert_date_to_iso() -> None:
    assert convert_date_to_iso(datetime.date(2025, 12, 17)) == "2025-12-17"


# ----------------------------------------------------------------------
# Default type encoders
# ----------------------------------------------------------------------


def test_default_datetime_encoder() -> None:
    encoder = DEFAULT_TYPE_ENCODERS[datetime.datetime]
    assert encoder(datetime.datetime(2025, 12, 17, 10, 30, 0)) == "2025-12-17T10:30:00Z"


def test_default_date_encoder() -> None:
    encoder = DEFAULT_TYPE_ENCODERS[datetime.date]
    assert encoder(datetime.date(2025, 12, 17)) == "2025-12-17"


def test_default_time_encoder() -> None:
    encoder = DEFAULT_TYPE_ENCODERS[datetime.time]
    assert encoder(datetime.time(10, 30, 0)) == "10:30:00"


def test_default_timedelta_encoder() -> None:
    encoder = DEFAULT_TYPE_ENCODERS[datetime.timedelta]
    assert encoder(datetime.timedelta(hours=1, minutes=30)) == 5400.0


def test_default_decimal_encoder() -> None:
    encoder = DEFAULT_TYPE_ENCODERS[Decimal]
    assert encoder(Decimal("123.456")) == "123.456"


def test_default_uuid_encoder() -> None:
    u = UUID("12345678-1234-5678-1234-567812345678")
    encoder = DEFAULT_TYPE_ENCODERS[UUID]
    assert encoder(u) == "12345678-1234-5678-1234-567812345678"


def test_default_path_encoder() -> None:
    encoder = DEFAULT_TYPE_ENCODERS[Path]
    assert encoder(Path("/tmp/test.txt")) == "/tmp/test.txt"


def test_default_pure_path_encoder() -> None:
    encoder = DEFAULT_TYPE_ENCODERS[PurePath]
    assert encoder(PurePath("/tmp/test.txt")) == "/tmp/test.txt"


def test_default_enum_encoder() -> None:
    encoder = DEFAULT_TYPE_ENCODERS[Enum]
    assert encoder(MyEnum.VALUE_A) == "a"
    assert encoder(MyEnum.VALUE_B) == "b"


def test_default_bytes_encoder() -> None:
    encoder = DEFAULT_TYPE_ENCODERS[bytes]
    assert encoder(b"hello") == "hello"


def test_default_set_and_frozenset_encoders() -> None:
    assert DEFAULT_TYPE_ENCODERS[set]({1, 2, 3}) == [1, 2, 3]
    assert DEFAULT_TYPE_ENCODERS[frozenset](frozenset({"a"})) == ["a"]


@pytest.mark.skipif(not PYDANTIC_INSTALLED, reason="Pydantic not installed")
def test_schema_dump_config_pydantic_exclude_unset_false_includes_defaults() -> None:
    """SchemaDumpConfig should pass exclude_unset through to Pydantic model_dump."""
    from pydantic import BaseModel

    class UpdateSchema(BaseModel):
        name: str
        is_admin: bool = False

    result = schema_dump(UpdateSchema(name="Ada"), config=SchemaDumpConfig(exclude_unset=False))

    assert result == {"name": "Ada", "is_admin": False}


@pytest.mark.skipif(not PYDANTIC_INSTALLED, reason="Pydantic not installed")
def test_schema_dump_config_pydantic_missing_sentinel_is_excluded_by_default() -> None:
    """Pydantic MISSING remains sentinel-like under the default dump policy."""
    from pydantic import BaseModel

    missing_module = pytest.importorskip("pydantic.experimental.missing_sentinel")

    class UpdateSchema(BaseModel):
        name: str = "Ada"
        marker: Any = missing_module.MISSING

    result = schema_dump(UpdateSchema(), config=SchemaDumpConfig(exclude_unset=False))

    assert result == {"name": "Ada"}


def test_pydantic_missing_sentinel_is_unavailable_when_pydantic_support_is_disabled(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Pydantic MISSING lookup should be skipped when Pydantic support is unavailable."""
    monkeypatch.setattr("advanced_alchemy.utils.serialization.PYDANTIC_INSTALLED", False)

    assert serialization._get_pydantic_missing_sentinel() is None


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
def test_schema_dump_config_msgspec_exclude_sentinels() -> None:
    """SchemaDumpConfig should control msgspec UNSET sentinel filtering."""
    import msgspec

    class UpdateStruct(msgspec.Struct):
        name: Any = msgspec.UNSET
        is_admin: bool = False
        notes: Any = None

    data = UpdateStruct()

    assert schema_dump(data) == {"is_admin": False, "notes": None}
    assert schema_dump(data, config=SchemaDumpConfig(exclude_sentinels=False)) == {
        "name": msgspec.UNSET,
        "is_admin": False,
        "notes": None,
    }
    assert schema_dump(data, config=SchemaDumpConfig(exclude_defaults=True, exclude_sentinels=False)) == {}
    assert schema_dump(data, config=SchemaDumpConfig(exclude_none=True)) == {"is_admin": False}


def test_msgspec_struct_dump_skips_declared_fields_without_values() -> None:
    """msgspec-style declared fields without values should be skipped."""

    class UpdateStruct:
        __struct_fields__ = ("name",)
        __struct_defaults__ = ()

    data: Any = UpdateStruct()

    assert serialization._dump_msgspec_struct(data, SchemaDumpConfig()) == {}


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
def test_schema_dump_exclude_unset_false_includes_msgspec_unset() -> None:
    """exclude_unset=False should include msgspec UNSET values for existing callers."""
    import msgspec

    class UpdateStruct(msgspec.Struct):
        name: Any = msgspec.UNSET

    assert schema_dump(UpdateStruct(), exclude_unset=False) == {"name": msgspec.UNSET}


@pytest.mark.skipif(not PYDANTIC_INSTALLED, reason="Pydantic not installed")
def test_schema_dump_config_exclude_unset_false_filters_pydantic_missing_by_default() -> None:
    """SchemaDumpConfig(exclude_unset=False) should filter sentinels by default."""
    from pydantic import BaseModel

    missing_module = pytest.importorskip("pydantic.experimental.missing_sentinel")

    class UpdateSchema(BaseModel):
        marker: Any = missing_module.MISSING

    assert schema_dump(UpdateSchema(), config=SchemaDumpConfig(exclude_unset=False)) == {}


@pytest.mark.skipif(not PYDANTIC_INSTALLED, reason="Pydantic not installed")
def test_schema_dump_config_pydantic_can_include_missing_sentinel() -> None:
    """SchemaDumpConfig should be able to include Pydantic MISSING when requested."""
    from pydantic import BaseModel

    missing_module = pytest.importorskip("pydantic.experimental.missing_sentinel")

    class UpdateSchema(BaseModel):
        name: str = "Ada"
        marker: Any = missing_module.MISSING

    assert schema_dump(
        UpdateSchema(),
        config=SchemaDumpConfig(exclude_unset=False, exclude_sentinels=False),
    ) == {"name": "Ada", "marker": missing_module.MISSING}


def test_schema_dump_config_dataclass_exclude_sentinels() -> None:
    """SchemaDumpConfig should apply Advanced Alchemy Empty filtering to dataclasses."""
    from dataclasses import dataclass

    @dataclass
    class UpdateDataclass:
        name: str
        is_admin: bool = False
        notes: Any = None
        marker: Any = Empty

    data = UpdateDataclass(name="Ada")

    assert schema_dump(data) == {"name": "Ada", "is_admin": False, "notes": None}
    assert schema_dump(data, config=SchemaDumpConfig(exclude_sentinels=False)) == {
        "name": "Ada",
        "is_admin": False,
        "notes": None,
        "marker": Empty,
    }
    assert schema_dump(data, config=SchemaDumpConfig(exclude_defaults=True, exclude_sentinels=False)) == {"name": "Ada"}
    assert schema_dump(data, config=SchemaDumpConfig(exclude_none=True)) == {"name": "Ada", "is_admin": False}


def test_schema_dump_config_dataclass_nested_instances() -> None:
    """Nested dataclass instances should be dumped recursively."""
    from dataclasses import dataclass

    @dataclass
    class ChildDataclass:
        name: str

    @dataclass
    class ParentDataclass:
        child: ChildDataclass
        is_admin: bool = False

    assert schema_dump(ParentDataclass(child=ChildDataclass(name="Ada"))) == {
        "child": {"name": "Ada"},
        "is_admin": False,
    }


@pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
def test_schema_dump_config_attrs_exclude_defaults_and_none() -> None:
    """SchemaDumpConfig should apply default and None filtering to attrs instances."""
    from attrs import define

    @define
    class UpdateAttrs:
        name: str
        is_admin: bool = False
        notes: Any = None

    data = UpdateAttrs(name="Ada")

    assert schema_dump(data, config=SchemaDumpConfig(exclude_defaults=True)) == {"name": "Ada"}
    assert schema_dump(data, config=SchemaDumpConfig(exclude_none=True)) == {"name": "Ada", "is_admin": False}


@pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
def test_schema_dump_config_attrs_exclude_sentinels() -> None:
    """SchemaDumpConfig should filter attrs NOTHING sentinel values."""
    from attrs import NOTHING, define

    @define
    class UpdateAttrs:
        name: str
        marker: Any

    data = UpdateAttrs(name="Ada", marker=NOTHING)

    assert schema_dump(data) == {"name": "Ada"}
    assert schema_dump(data, config=SchemaDumpConfig(exclude_sentinels=False, exclude_none=True)) == {
        "name": "Ada",
        "marker": NOTHING,
    }


@pytest.mark.skipif(
    not ATTRS_INSTALLED or not serialization.CATTRS_INSTALLED,
    reason="attrs and cattrs are not installed",
)
def test_schema_dump_config_attrs_without_filtering_uses_cattrs() -> None:
    """attrs instances without active filters should use cattrs when it is available."""
    from attrs import define

    @define
    class UpdateAttrs:
        name: str
        is_admin: bool = False

    assert schema_dump(
        UpdateAttrs(name="Ada"),
        config=SchemaDumpConfig(exclude_sentinels=False),
    ) == {"name": "Ada", "is_admin": False}


@pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
def test_schema_dump_config_attrs_without_cattrs_uses_attrs_asdict(monkeypatch: pytest.MonkeyPatch) -> None:
    """attrs instances should still dump when cattrs is unavailable."""
    from attrs import define

    @define
    class UpdateAttrs:
        name: str
        is_admin: bool = False

    monkeypatch.setattr("advanced_alchemy.utils.serialization.CATTRS_INSTALLED", False)

    assert schema_dump(
        UpdateAttrs(name="Ada"),
        config=SchemaDumpConfig(exclude_sentinels=False),
    ) == {"name": "Ada", "is_admin": False}


def test_schema_dump_skips_pydantic_branch_when_pydantic_not_installed(monkeypatch: pytest.MonkeyPatch) -> None:
    """schema_dump should not rely on Pydantic APIs when Pydantic is unavailable."""

    class PydanticLike:
        def __init__(self) -> None:
            self.name = "Ada"

        def model_dump(self, **kwargs: Any) -> dict[str, Any]:
            msg = "model_dump should not be called when Pydantic is unavailable"
            raise AssertionError(msg)

    monkeypatch.setattr("advanced_alchemy.utils.serialization.PYDANTIC_INSTALLED", False)

    assert schema_dump(PydanticLike()) == {"name": "Ada"}


def test_schema_dump_skips_msgspec_branch_when_msgspec_not_installed(monkeypatch: pytest.MonkeyPatch) -> None:
    """schema_dump should not rely on msgspec APIs when msgspec is unavailable."""

    class MsgspecLike:
        __struct_fields__ = ("name",)

        def __init__(self) -> None:
            self.name = "Ada"

    monkeypatch.setattr("advanced_alchemy.utils.serialization.MSGSPEC_INSTALLED", False)

    assert schema_dump(MsgspecLike()) == {"name": "Ada"}


def test_schema_dump_skips_attrs_branch_when_attrs_not_installed(monkeypatch: pytest.MonkeyPatch) -> None:
    """schema_dump should not rely on attrs APIs when attrs is unavailable."""

    class AttrsLike:
        __attrs_attrs__ = ()

        def __init__(self) -> None:
            self.name = "Ada"

    monkeypatch.setattr("advanced_alchemy.utils.serialization.ATTRS_INSTALLED", False)

    assert schema_dump(AttrsLike()) == {"name": "Ada"}


@pytest.mark.parametrize(
    "value",
    [
        None,
        1,
        [1, 2],
        {},
    ],
)
def test_has_dict_attribute_rejects_values_without_instance_dict(value: Any) -> None:
    assert not serialization.has_dict_attribute(value)


def test_has_dict_attribute_accepts_plain_object() -> None:
    class PlainObject:
        def __init__(self) -> None:
            self.name = "Ada"

    assert serialization.has_dict_attribute(PlainObject())


def test_has_dict_attribute_rejects_slots_only_object() -> None:
    class SlotsOnlyObject:
        __slots__ = ("name",)

        def __init__(self) -> None:
            self.name = "Ada"

    assert not serialization.has_dict_attribute(SlotsOnlyObject())


def test_schema_dump_uses_instance_dict_fallback_for_plain_objects() -> None:
    class PlainObject:
        def __init__(self) -> None:
            self.name = "Ada"

    assert schema_dump(PlainObject()) == {"name": "Ada"}


def test_schema_dump_passes_through_slots_only_objects() -> None:
    class SlotsOnlyObject:
        __slots__ = ("name",)

        def __init__(self) -> None:
            self.name = "Ada"

    value = SlotsOnlyObject()
    result: Any = schema_dump(value)

    assert result is value


def test_serialization_module_does_not_import_public_dict_protocol() -> None:
    assert "DictProtocol" not in vars(serialization)


@pytest.mark.parametrize(
    ("kind", "value", "expected"),
    [
        (IPv4Address, "192.168.1.1", "192.168.1.1"),
        (IPv4Interface, "192.168.1.1/24", "192.168.1.1/24"),
        (IPv4Network, "192.168.1.0/24", "192.168.1.0/24"),
        (IPv6Address, "::1", "::1"),
        (IPv6Interface, "::1/128", "::1/128"),
        (IPv6Network, "::/0", "::/0"),
    ],
)
def test_default_ip_address_encoders(kind: type, value: str, expected: str) -> None:
    encoder = DEFAULT_TYPE_ENCODERS[kind]
    assert encoder(kind(value)) == expected


# ----------------------------------------------------------------------
# Msgspec serializer
# ----------------------------------------------------------------------


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
def test_msgspec_encode_simple_dict() -> None:
    result = MsgspecSerializer().encode({"key": "value", "number": 42})
    assert result == '{"key":"value","number":42}'


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
def test_msgspec_encode_as_bytes() -> None:
    result = MsgspecSerializer().encode({"key": "value"}, as_bytes=True)
    assert isinstance(result, bytes)
    assert result == b'{"key":"value"}'


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
def test_msgspec_decode_string() -> None:
    assert MsgspecSerializer().decode('{"key":"value","number":42}') == {"key": "value", "number": 42}


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
def test_msgspec_decode_bytes() -> None:
    assert MsgspecSerializer().decode(b'{"key":"value"}') == {"key": "value"}


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
def test_msgspec_decode_bytes_passthrough() -> None:
    data = b'{"key":"value"}'
    assert MsgspecSerializer().decode(data, decode_bytes=False) == data


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
def test_msgspec_custom_type_encoders() -> None:
    custom_encoders: TypeEncodersMap = {CustomType: lambda x: {"custom_value": x.value}}
    serializer = MsgspecSerializer(type_encoders=custom_encoders)
    assert serializer.encode({"obj": CustomType("test")}) == '{"obj":{"custom_value":"test"}}'


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
def test_msgspec_custom_enc_hook() -> None:
    def my_hook(obj: Any) -> Any:
        if isinstance(obj, CustomType):
            return f"custom:{obj.value}"
        raise TypeError

    serializer = MsgspecSerializer(enc_hook=my_hook)
    assert serializer.encode({"obj": CustomType("test")}) == '{"obj":"custom:test"}'


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
def test_msgspec_mro_lookup() -> None:
    custom_encoders: TypeEncodersMap = {CustomType: lambda x: f"parent:{x.value}"}
    serializer = MsgspecSerializer(type_encoders=custom_encoders)
    assert serializer.encode({"obj": ChildCustomType("child")}) == '{"obj":"parent:child"}'


@pytest.mark.skipif(not MSGSPEC_INSTALLED, reason="msgspec not installed")
def test_msgspec_custom_type_encoders_override_native_values() -> None:
    serializer = MsgspecSerializer(
        type_encoders={
            Decimal: lambda _: "custom-decimal",
            UUID: lambda _: "custom-uuid",
            bytes: lambda _: "custom-bytes",
            datetime.datetime: lambda _: "custom-datetime",
        }
    )

    result = json.loads(
        serializer.encode(
            {
                "decimal": Decimal("1.2"),
                "uuid": UUID("12345678-1234-5678-1234-567812345678"),
                "bytes": b"abc",
                "datetime": datetime.datetime(2026, 1, 2, 3, 4, 5, tzinfo=datetime.timezone.utc),
            }
        )
    )

    assert result == {
        "decimal": "custom-decimal",
        "uuid": "custom-uuid",
        "bytes": "custom-bytes",
        "datetime": "custom-datetime",
    }


# ----------------------------------------------------------------------
# Orjson serializer
# ----------------------------------------------------------------------


@pytest.mark.skipif(not ORJSON_INSTALLED, reason="orjson not installed")
def test_orjson_encode_simple_dict() -> None:
    assert OrjsonSerializer().encode({"key": "value", "number": 42}) == '{"key":"value","number":42}'


@pytest.mark.skipif(not ORJSON_INSTALLED, reason="orjson not installed")
def test_orjson_encode_as_bytes() -> None:
    result = OrjsonSerializer().encode({"key": "value"}, as_bytes=True)
    assert isinstance(result, bytes)
    assert result == b'{"key":"value"}'


@pytest.mark.skipif(not ORJSON_INSTALLED, reason="orjson not installed")
def test_orjson_decode_string() -> None:
    assert OrjsonSerializer().decode('{"key":"value","number":42}') == {"key": "value", "number": 42}


@pytest.mark.skipif(not ORJSON_INSTALLED, reason="orjson not installed")
def test_orjson_decode_bytes() -> None:
    assert OrjsonSerializer().decode(b'{"key":"value"}') == {"key": "value"}


@pytest.mark.skipif(not ORJSON_INSTALLED, reason="orjson not installed")
def test_orjson_decode_bytes_passthrough() -> None:
    data = b'{"key":"value"}'
    assert OrjsonSerializer().decode(data, decode_bytes=False) == data


@pytest.mark.skipif(not ORJSON_INSTALLED, reason="orjson not installed")
def test_orjson_custom_type_encoders() -> None:
    custom_encoders: TypeEncodersMap = {CustomType: lambda x: {"custom_value": x.value}}
    serializer = OrjsonSerializer(type_encoders=custom_encoders)
    assert serializer.encode({"obj": CustomType("test")}) == '{"obj":{"custom_value":"test"}}'


@pytest.mark.skipif(not ORJSON_INSTALLED, reason="orjson not installed")
def test_orjson_mro_lookup() -> None:
    custom_encoders: TypeEncodersMap = {CustomType: lambda x: f"parent:{x.value}"}
    serializer = OrjsonSerializer(type_encoders=custom_encoders)
    parsed = json.loads(serializer.encode({"obj": ChildCustomType("child")}))
    assert parsed["obj"] == "parent:child"


@pytest.mark.skipif(not ORJSON_INSTALLED, reason="orjson not installed")
@pytest.mark.skipif(not NUMPY_INSTALLED, reason="numpy not installed")
def test_orjson_uses_numpy_option() -> None:
    """Cover the OPT_SERIALIZE_NUMPY branch in OrjsonSerializer.encode."""
    import numpy as np

    assert OrjsonSerializer().encode({"arr": np.array([1, 2, 3])}) == '{"arr":[1,2,3]}'


# ----------------------------------------------------------------------
# Standard library serializer
# ----------------------------------------------------------------------


def test_stdlib_encode_simple_dict() -> None:
    assert StandardLibSerializer().encode({"key": "value", "number": 42}) == '{"key": "value", "number": 42}'


def test_stdlib_encode_as_bytes() -> None:
    result = StandardLibSerializer().encode({"key": "value"}, as_bytes=True)
    assert isinstance(result, bytes)
    assert result == b'{"key": "value"}'


def test_stdlib_decode_string() -> None:
    assert StandardLibSerializer().decode('{"key": "value", "number": 42}') == {"key": "value", "number": 42}


def test_stdlib_decode_bytes() -> None:
    assert StandardLibSerializer().decode(b'{"key": "value"}') == {"key": "value"}


def test_stdlib_decode_bytes_passthrough() -> None:
    data = b'{"key":"value"}'
    assert StandardLibSerializer().decode(data, decode_bytes=False) == data


def test_stdlib_custom_type_encoders() -> None:
    custom_encoders: TypeEncodersMap = {CustomType: lambda x: {"custom_value": x.value}}
    serializer = StandardLibSerializer(type_encoders=custom_encoders)
    assert serializer.encode({"obj": CustomType("test")}) == '{"obj": {"custom_value": "test"}}'


def test_stdlib_mro_lookup() -> None:
    custom_encoders: TypeEncodersMap = {CustomType: lambda x: f"parent:{x.value}"}
    serializer = StandardLibSerializer(type_encoders=custom_encoders)
    assert serializer.encode({"obj": ChildCustomType("child")}) == '{"obj": "parent:child"}'


# ----------------------------------------------------------------------
# Encoding-hook fallback inside _create_enc_hook
# ----------------------------------------------------------------------


def test_enc_hook_falls_back_to_str_for_unknown_types() -> None:
    """Types with no encoder use str() so msgspec/orjson don't raise."""
    hook = StandardLibSerializer()._create_enc_hook()

    class Plain:
        def __str__(self) -> str:
            return "plain-repr"

    assert hook(Plain()) == "plain-repr"


def test_enc_hook_raises_type_error_when_str_fails() -> None:
    """When str() itself raises, _create_enc_hook reports TypeError."""
    hook = StandardLibSerializer()._create_enc_hook()

    class Hostile:
        def __str__(self) -> str:
            msg = "boom"
            raise RuntimeError(msg)

    with pytest.raises(TypeError, match="Cannot serialize Hostile"):
        hook(Hostile())


# ----------------------------------------------------------------------
# Serializer factory
# ----------------------------------------------------------------------


def test_get_serializer_default_is_singleton() -> None:
    s1 = get_serializer()
    s2 = get_serializer()
    assert s1 is s2
    assert isinstance(s1, JSONSerializer)


def test_get_serializer_caches_by_configuration() -> None:
    encoders: TypeEncodersMap = {CustomType: lambda x: x.value}
    s1 = get_serializer(encoders)
    s2 = get_serializer(encoders)
    assert s1 is s2

    s3 = get_serializer({CustomType: lambda x: "other"})
    assert s3 is not s1


def test_get_serializer_caches_repeated_lookups() -> None:
    """A second call with the same encoders hits the cache branch."""
    encoders: TypeEncodersMap = {Decimal: str}
    a = get_serializer(encoders)
    b = get_serializer(encoders)
    assert a is b


def test_get_serializer_thread_safety() -> None:
    results: list[JSONSerializer] = []

    def worker() -> None:
        results.append(get_serializer())

    threads = [threading.Thread(target=worker) for _ in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert len(results) == 10
    assert all(r is results[0] for r in results)


def test_get_serializer_default_initialization_is_locked(monkeypatch: pytest.MonkeyPatch) -> None:
    import advanced_alchemy.utils.serialization as serialization_module

    created: list[JSONSerializer] = []
    creation_lock = threading.Lock()
    start_barrier = threading.Barrier(20)
    monkeypatch.setattr(serialization_module, "_default_serializer", None)

    def create_default_serializer() -> JSONSerializer:
        time.sleep(0.01)
        serializer = StandardLibSerializer()
        with creation_lock:
            created.append(serializer)
        return serializer

    monkeypatch.setattr(serialization_module, "_create_default_serializer", create_default_serializer)

    results: list[JSONSerializer] = []

    def worker() -> None:
        start_barrier.wait()
        results.append(serialization_module.get_serializer())

    threads = [threading.Thread(target=worker) for _ in range(20)]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()

    assert len(created) == 1
    assert len(results) == 20
    assert all(result is created[0] for result in results)


# ----------------------------------------------------------------------
# Public API
# ----------------------------------------------------------------------


def test_encode_json_default() -> None:
    result = encode_json({"key": "value"})
    assert isinstance(result, str)
    assert "key" in result


def test_encode_json_with_type_encoders() -> None:
    result = encode_json({"obj": CustomType("test")}, type_encoders={CustomType: lambda x: x.value})
    assert result == '{"obj":"test"}'


def test_encode_json_custom_encoder_overrides_backend_native_type() -> None:
    result = encode_json({"x": Decimal("1.2")}, type_encoders={Decimal: lambda _: "custom"})
    assert result == '{"x":"custom"}'


def test_encode_json_as_bytes() -> None:
    result = encode_json({"key": "value"}, as_bytes=True)
    assert isinstance(result, bytes)
    assert b"key" in result


def test_decode_json() -> None:
    assert decode_json('{"key":"value"}') == {"key": "value"}


def test_decode_json_bytes() -> None:
    assert decode_json(b'{"key":"value"}') == {"key": "value"}


# ----------------------------------------------------------------------
# Backward compatibility helpers
# ----------------------------------------------------------------------


def test_all_exports_present() -> None:
    from advanced_alchemy.utils.serialization import __all__

    expected = {
        "BaseModel",
        "DEFAULT_TYPE_ENCODERS",
        "FilterTypeT",
        "JSONSerializer",
        "ModelDTOT",
        "ModelDictListT",
        "ModelDictT",
        "MsgspecSerializer",
        "OrjsonSerializer",
        "StandardLibSerializer",
        "SupportedSchemaModel",
        "TypeEncodersMap",
        "convert_date_to_iso",
        "convert_datetime_to_gmt_iso",
        "decode_complex_type",
        "decode_json",
        "encode_complex_type",
        "encode_json",
        "get_serializer",
        "is_schema",
        "schema_dump",
    }
    assert expected.issubset(set(__all__))


def test_encode_complex_type_datetime() -> None:
    dt = datetime.datetime(2025, 12, 17, 10, 30, 0)
    assert encode_complex_type(dt) == {"__type__": "datetime", "value": "2025-12-17T10:30:00"}


def test_encode_complex_type_date() -> None:
    d = datetime.date(2025, 12, 17)
    assert encode_complex_type(d) == {"__type__": "date", "value": "2025-12-17"}


def test_encode_complex_type_time() -> None:
    t = datetime.time(10, 30, 0)
    assert encode_complex_type(t) == {"__type__": "time", "value": "10:30:00"}


def test_encode_complex_type_timedelta() -> None:
    td = datetime.timedelta(seconds=42)
    assert encode_complex_type(td) == {"__type__": "timedelta", "value": 42.0}


def test_encode_complex_type_decimal() -> None:
    assert encode_complex_type(Decimal("19.99")) == {"__type__": "decimal", "value": "19.99"}


def test_encode_complex_type_bytes() -> None:
    assert encode_complex_type(b"\x00\xff") == {"__type__": "bytes", "value": "00ff"}


def test_encode_complex_type_uuid() -> None:
    u = UUID("12345678-1234-5678-1234-567812345678")
    assert encode_complex_type(u) == {"__type__": "uuid", "value": str(u)}


def test_encode_complex_type_set_and_frozenset() -> None:
    result = encode_complex_type({1, 2})
    assert isinstance(result, dict) and result["__type__"] == "set"
    assert sorted(result["value"]) == [1, 2]
    result = encode_complex_type(frozenset({3}))
    assert isinstance(result, dict) and result["__type__"] == "set"
    assert result["value"] == [3]


def test_encode_complex_type_unsupported_returns_none() -> None:
    assert encode_complex_type(object()) is None


def test_decode_complex_type_passes_through_primitives() -> None:
    assert decode_complex_type(42) == 42
    assert decode_complex_type("text") == "text"


def test_decode_complex_type_recurses_into_lists() -> None:
    encoded = [{"__type__": "uuid", "value": "12345678-1234-5678-1234-567812345678"}, 7]
    decoded = decode_complex_type(encoded)
    assert isinstance(decoded[0], UUID)
    assert decoded[1] == 7


def test_decode_complex_type_passes_through_plain_dict() -> None:
    assert decode_complex_type({"x": 1, "y": 2}) == {"x": 1, "y": 2}


def test_decode_complex_type_unknown_marker_returns_dict() -> None:
    marker = {"__type__": "mystery", "value": "anything"}
    assert decode_complex_type(marker) == marker


@pytest.mark.parametrize(
    ("marker", "expected"),
    [
        ({"__type__": "datetime", "value": "2025-12-17T10:30:00"}, datetime.datetime(2025, 12, 17, 10, 30, 0)),
        ({"__type__": "date", "value": "2025-12-17"}, datetime.date(2025, 12, 17)),
        ({"__type__": "time", "value": "10:30:00"}, datetime.time(10, 30, 0)),
        ({"__type__": "timedelta", "value": 42.0}, datetime.timedelta(seconds=42)),
        ({"__type__": "decimal", "value": "19.99"}, Decimal("19.99")),
        ({"__type__": "bytes", "value": "00ff"}, b"\x00\xff"),
        (
            {"__type__": "uuid", "value": "12345678-1234-5678-1234-567812345678"},
            UUID("12345678-1234-5678-1234-567812345678"),
        ),
        ({"__type__": "set", "value": [1, 2]}, {1, 2}),
    ],
)
def test_decode_complex_type_each_typed_marker(marker: dict[str, Any], expected: Any) -> None:
    assert decode_complex_type(marker) == expected


def _import_attr(module_name: str, attr_name: str) -> Any:
    module = importlib.import_module(module_name)
    return getattr(module, attr_name)


@pytest.mark.parametrize(
    ("name", "new_module"),
    [
        ("BaseModel", "advanced_alchemy.typing"),
        ("PYDANTIC_INSTALLED", "advanced_alchemy.typing"),
        ("encode_json", "advanced_alchemy.utils.serialization"),
        ("decode_json", "advanced_alchemy.utils.serialization"),
        ("encode_complex_type", "advanced_alchemy.utils.serialization"),
        ("decode_complex_type", "advanced_alchemy.utils.serialization"),
        ("convert_datetime_to_gmt_iso", "advanced_alchemy.utils.serialization"),
        ("convert_date_to_iso", "advanced_alchemy.utils.serialization"),
    ],
)
def test_legacy_serialization_shim_emits_deprecation_warning(name: str, new_module: str) -> None:
    with pytest.warns(DeprecationWarning, match=rf"Use '{new_module}\.{name}' instead"):
        value = _import_attr("advanced_alchemy._serialization", name)

    canonical = getattr(importlib.import_module(new_module), name)
    assert value is canonical


def test_legacy_serialization_shim_unknown_attribute_raises() -> None:
    with pytest.raises(AttributeError, match="not_a_real_name"):
        _import_attr("advanced_alchemy._serialization", "not_a_real_name")


def test_legacy_serialization_shim_lists_renames_in_all() -> None:
    import advanced_alchemy._serialization as shim

    expected = {
        "BaseModel",
        "PYDANTIC_INSTALLED",
        "encode_json",
        "decode_json",
        "encode_complex_type",
        "decode_complex_type",
        "convert_datetime_to_gmt_iso",
        "convert_date_to_iso",
    }
    assert expected.issubset(set(shim.__all__))


@pytest.mark.parametrize(
    ("name", "new_module"),
    [
        # Foundational typing surface
        ("ATTRS_INSTALLED", "advanced_alchemy.typing"),
        ("CATTRS_INSTALLED", "advanced_alchemy.typing"),
        ("LITESTAR_INSTALLED", "advanced_alchemy.typing"),
        ("MSGSPEC_INSTALLED", "advanced_alchemy.typing"),
        ("PYDANTIC_INSTALLED", "advanced_alchemy.typing"),
        ("SQLMODEL_INSTALLED", "advanced_alchemy.typing"),
        ("UNSET", "advanced_alchemy.typing"),
        ("BaseModel", "advanced_alchemy.typing"),
        ("Struct", "advanced_alchemy.typing"),
        ("TypeAdapter", "advanced_alchemy.typing"),
        ("UnsetType", "advanced_alchemy.typing"),
        ("FailFast", "advanced_alchemy.typing"),
        ("DTOData", "advanced_alchemy.typing"),
        ("T", "advanced_alchemy.typing"),
        # Schema/dict/attrs guards and helpers
        ("schema_dump", "advanced_alchemy.utils.serialization"),
        ("ModelDictT", "advanced_alchemy.utils.serialization"),
        ("ModelDictListT", "advanced_alchemy.utils.serialization"),
        ("ModelDTOT", "advanced_alchemy.utils.serialization"),
        ("FilterTypeT", "advanced_alchemy.utils.serialization"),
        ("SupportedSchemaModel", "advanced_alchemy.utils.serialization"),
        ("is_attrs_instance", "advanced_alchemy.utils.serialization"),
        ("is_dto_data", "advanced_alchemy.utils.serialization"),
        ("is_msgspec_struct", "advanced_alchemy.utils.serialization"),
        ("is_pydantic_model", "advanced_alchemy.utils.serialization"),
        ("is_schema", "advanced_alchemy.utils.serialization"),
        ("is_dict", "advanced_alchemy.utils.serialization"),
        ("get_type_adapter", "advanced_alchemy.utils.serialization"),
        ("structure", "advanced_alchemy.utils.serialization"),
        ("unstructure", "advanced_alchemy.utils.serialization"),
        ("fields", "advanced_alchemy.utils.serialization"),
        ("asdict", "advanced_alchemy.utils.serialization"),
        ("has", "advanced_alchemy.utils.serialization"),
    ],
)
def test_legacy_service_typing_shim_emits_deprecation_warning(name: str, new_module: str) -> None:
    with pytest.warns(DeprecationWarning, match=rf"Use '{new_module}\.{name}' instead"):
        value = _import_attr("advanced_alchemy.service.typing", name)

    canonical = getattr(importlib.import_module(new_module), name)
    assert value is canonical


def test_typing_dict_protocol_emits_deprecation_warning() -> None:
    with pytest.warns(DeprecationWarning, match=r"DictProtocol.*1\.11\.0.*removed in 2\.0\.0"):
        value = _import_attr("advanced_alchemy.typing", "DictProtocol")

    assert value is _import_attr("advanced_alchemy._typing", "DictProtocol")


def test_typing_canonical_exports_do_not_emit_deprecation_warning() -> None:
    typing_module = importlib.import_module("advanced_alchemy.typing")

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert getattr(typing_module, "PYDANTIC_INSTALLED") is PYDANTIC_INSTALLED


def test_legacy_service_typing_dict_protocol_remains_compatible() -> None:
    with pytest.warns(DeprecationWarning, match="DictProtocol"):
        value = _import_attr("advanced_alchemy.service.typing", "DictProtocol")

    assert value is _import_attr("advanced_alchemy._typing", "DictProtocol")


def test_legacy_service_typing_shim_unknown_attribute_raises() -> None:
    with pytest.raises(AttributeError, match="not_a_real_name"):
        _import_attr("advanced_alchemy.service.typing", "not_a_real_name")


def test_legacy_service_typing_shim_keeps_pydantic_use_failfast() -> None:
    import advanced_alchemy.service.typing as shim

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert shim.PYDANTIC_USE_FAILFAST is False


# ----------------------------------------------------------------------
# Optional dependency integrations
# ----------------------------------------------------------------------


@pytest.mark.skipif(not NUMPY_INSTALLED, reason="numpy not installed")
def test_numpy_support() -> None:
    import numpy as np

    data = {"array": np.array([1, 2, 3]), "int": np.int64(42)}
    assert encode_json(data) == '{"array":[1,2,3],"int":42}'


@pytest.mark.skipif(not PYDANTIC_INSTALLED, reason="pydantic not installed")
def test_pydantic_support() -> None:
    from pydantic import BaseModel

    class MyModel(BaseModel):
        name: str
        age: int

    assert encode_json({"model": MyModel(name="test", age=25)}) == '{"model":{"name":"test","age":25}}'



"""SQLAlchemy filter constructs for advanced query operations.

This module provides a comprehensive collection of filter datastructures designed to
enhance SQLAlchemy query construction. It implements type-safe, reusable filter patterns
for common database query operations.

Features:
    Type-safe filter construction, datetime range filtering, collection-based filtering,
    pagination support, search operations, and customizable ordering.

Note:
    All filter classes implement the :class:`StatementFilter` ABC, ensuring consistent
    interface across different filter types.

See Also:
    - :class:`sqlalchemy.sql.expression.Select`: Core SQLAlchemy select expression
    - :class:`sqlalchemy.orm.Query`: SQLAlchemy ORM query interface
    - :mod:`advanced_alchemy.base`: Base model definitions

"""

import datetime
import logging
from abc import ABC, abstractmethod
from collections.abc import Collection
from dataclasses import dataclass
from operator import attrgetter
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Generic,
    Literal,
    Optional,
    Union,
    cast,
)

from sqlalchemy import (
    BinaryExpression,
    ColumnElement,
    Date,
    Delete,
    Select,
    Update,
    and_,
    any_,
    exists,
    false,
    not_,
    or_,
    select,
    text,
    true,
)
from sqlalchemy.sql import operators as op
from sqlalchemy.sql.dml import ReturningDelete, ReturningUpdate
from typing_extensions import TypeAlias, TypedDict, TypeVar

from advanced_alchemy.base import ModelProtocol

if TYPE_CHECKING:
    from sqlalchemy.orm import InstrumentedAttribute

    FilterFieldName: TypeAlias = Union[str, ColumnElement[Any], InstrumentedAttribute[Any]]
    InstrumentedField: TypeAlias = Union[ColumnElement[Any], InstrumentedAttribute[Any]]
else:
    FilterFieldName: TypeAlias = Any
    InstrumentedField: TypeAlias = Any

__all__ = (
    "BeforeAfter",
    "BooleanFilter",
    "ChoicesFilter",
    "CollectionFilter",
    "ComparisonFilter",
    "ExistsFilter",
    "FilterGroup",
    "FilterMap",
    "FilterTypes",
    "InAnyFilter",
    "LimitOffset",
    "LogicalOperatorMap",
    "MultiFilter",
    "NotExistsFilter",
    "NotInCollectionFilter",
    "NotInSearchFilter",
    "NotNullFilter",
    "NullFilter",
    "OnBeforeAfter",
    "OrderBy",
    "PaginationFilter",
    "SearchFilter",
    "StatementFilter",
    "StatementFilterT",
    "StatementTypeT",
)

T = TypeVar("T")
ModelT = TypeVar("ModelT", bound=ModelProtocol)
StatementFilterT = TypeVar("StatementFilterT", bound="StatementFilter")
StatementTypeT = TypeVar(
    "StatementTypeT",
    bound=Union[
        ReturningDelete[tuple[Any]], ReturningUpdate[tuple[Any]], Select[tuple[Any]], Select[Any], Update, Delete
    ],
)

logger = logging.getLogger("advanced_alchemy")


# Define TypedDicts for filter and logical maps
class FilterMap(TypedDict):
    before_after: "type[BeforeAfter]"
    on_before_after: "type[OnBeforeAfter]"
    boolean: "type[BooleanFilter]"
    choices: "type[ChoicesFilter[Any]]"
    collection: "type[CollectionFilter[Any]]"
    not_in_collection: "type[NotInCollectionFilter[Any]]"
    limit_offset: "type[LimitOffset]"
    null: "type[NullFilter]"
    not_null: "type[NotNullFilter]"
    order_by: "type[OrderBy]"
    search: "type[SearchFilter]"
    not_in_search: "type[NotInSearchFilter]"
    comparison: "type[ComparisonFilter]"
    exists: "type[ExistsFilter]"
    not_exists: "type[NotExistsFilter]"
    filter_group: "type[FilterGroup]"


class LogicalOperatorMap(TypedDict):
    and_: Callable[..., ColumnElement[bool]]
    or_: Callable[..., ColumnElement[bool]]


class StatementFilter(ABC):
    """Abstract base class for SQLAlchemy statement filters.

    This class defines the interface for all filter types in the system. Each filter
    implementation must provide a method to append its filtering logic to an existing
    SQLAlchemy statement.
    """

    @abstractmethod
    def append_to_statement(
        self, statement: StatementTypeT, model: type[ModelT], *args: Any, **kwargs: Any
    ) -> StatementTypeT:
        """Append filter conditions to a SQLAlchemy statement.

        Args:
            statement: The SQLAlchemy statement to modify
            model: The SQLAlchemy model class
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments

        Returns:
            StatementTypeT: Modified SQLAlchemy statement with filter conditions applied

        Raises:
            NotImplementedError: If the concrete class doesn't implement this method

        Note:
            This method must be implemented by all concrete filter classes.

        See Also:
            :meth:`sqlalchemy.sql.expression.Select.where`: SQLAlchemy where clause
        """
        return statement

    @staticmethod
    def _get_instrumented_attr(model: Any, key: FilterFieldName) -> InstrumentedField:
        """Get SQLAlchemy instrumented attribute from model.

        Args:
            model: SQLAlchemy model class or instance
            key: Attribute name or instrumented attribute

        Returns:
            InstrumentedAttribute[Any]: SQLAlchemy instrumented attribute

        See Also:
            :class:`sqlalchemy.orm.attributes.InstrumentedAttribute`: SQLAlchemy attribute
        """
        return cast("InstrumentedAttribute[Any]", getattr(model, key)) if isinstance(key, str) else key


@dataclass
class BeforeAfter(StatementFilter):
    """DateTime range filter with exclusive bounds.

    This filter creates date/time range conditions using < and > operators,
    excluding the boundary values.

    If either `before` or `after` is None, that boundary condition is not applied.

    See Also:
    ---------
        :class:`OnBeforeAfter` : Inclusive datetime range filtering

    """

    field_name: FilterFieldName
    """Field name, model attribute, or func expression."""
    before: Optional[datetime.datetime]
    """Filter results where field is earlier than this value."""
    after: Optional[datetime.datetime]
    """Filter results where field is later than this value."""

    def append_to_statement(self, statement: StatementTypeT, model: type[ModelT]) -> StatementTypeT:
        """Apply datetime range conditions to statement.

        Parameters
        ----------
        statement : StatementTypeT
            The SQLAlchemy statement to modify
        model : type[ModelT]
            The SQLAlchemy model class

        Returns:
        --------
        StatementTypeT
            Modified statement with datetime range conditions
        """
        field = self._get_instrumented_attr(model, self.field_name)
        if self.before is not None:
            statement = cast("StatementTypeT", statement.where(field < self.before))
        if self.after is not None:
            statement = cast("StatementTypeT", statement.where(field > self.after))
        return statement


@dataclass
class OnBeforeAfter(StatementFilter):
    """DateTime range filter with inclusive bounds.

    This filter creates date/time range conditions using <= and >= operators,
    including the boundary values.

    If either `on_or_before` or `on_or_after` is None, that boundary condition
    is not applied.

    See Also:
    ---------
        :class:`BeforeAfter` : Exclusive datetime range filtering

    """

    field_name: FilterFieldName
    """Field name, model attribute, or func expression."""
    on_or_before: Optional[datetime.datetime]
    """Filter results where field is on or earlier than this value."""
    on_or_after: Optional[datetime.datetime]
    """Filter results where field is on or later than this value."""

    def append_to_statement(self, statement: StatementTypeT, model: type[ModelT]) -> StatementTypeT:
        """Apply inclusive datetime range conditions to statement.

        Parameters
        ----------
        statement : StatementTypeT
            The SQLAlchemy statement to modify
        model : type[ModelT]
            The SQLAlchemy model class

        Returns:
        --------
        StatementTypeT
            Modified statement with inclusive datetime range conditions
        """
        field = self._get_instrumented_attr(model, self.field_name)
        if self.on_or_before is not None:
            statement = cast("StatementTypeT", statement.where(field <= self.on_or_before))
        if self.on_or_after is not None:
            statement = cast("StatementTypeT", statement.where(field >= self.on_or_after))
        return statement


class InAnyFilter(StatementFilter, ABC):
    """Base class for filters using IN or ANY operators.

    This abstract class provides common functionality for filters that check
    membership in a collection using either the SQL IN operator or the ANY operator.
    """


@dataclass
class CollectionFilter(InAnyFilter, Generic[T]):
    """Data required to construct a WHERE ... IN (...) clause.

    This filter restricts records based on a field's presence in a collection of values.

    The filter supports both ``IN`` and ``ANY`` operators for collection membership testing.
    Use ``prefer_any=True`` in ``append_to_statement`` to use the ``ANY`` operator.
    """

    field_name: FilterFieldName
    """Field name, model attribute, or func expression."""
    values: Union[Collection[T], None]
    """Values for the ``IN`` clause. If this is None, no filter is applied.
        An empty list will force an empty result set (WHERE 1=-1)"""

    def append_to_statement(
        self,
        statement: StatementTypeT,
        model: type[ModelT],
        prefer_any: bool = False,
    ) -> StatementTypeT:
        """Apply a WHERE ... IN or WHERE ... ANY (...) clause to the statement.

        Parameters
        ----------
        statement : StatementTypeT
            The SQLAlchemy statement to modify
        model : type[ModelT]
            The SQLAlchemy model class
        prefer_any : bool, optional
            If True, uses the SQLAlchemy :func:`any_` operator instead of
            :func:`in_` for the filter condition

        Returns:
        --------
        StatementTypeT
            Modified statement with the appropriate IN conditions
        """
        field = self._get_instrumented_attr(model, self.field_name)
        if self.values is None:
            return statement
        if not self.values:
            # Return empty result set by forcing a false condition
            return cast("StatementTypeT", statement.where(text("1=-1")))
        if prefer_any:
            return cast("StatementTypeT", statement.where(any_(self.values) == field))  # type: ignore[arg-type]
        return cast("StatementTypeT", statement.where(field.in_(self.values)))


@dataclass
class ChoicesFilter(CollectionFilter[T]):
    """Filter records whose field matches one of a fixed set of choices."""


@dataclass
class BooleanFilter(StatementFilter):
    """Filter records by a boolean field value.

    If ``value`` is ``None``, the statement is returned unchanged.
    """

    field_name: "Union[str, ColumnElement[Any], InstrumentedAttribute[Any]]"
    """Field name, model attribute, or func expression."""
    value: Optional[bool]
    """Boolean value to match. If None, no filter is applied."""

    def append_to_statement(self, statement: StatementTypeT, model: type[ModelT]) -> StatementTypeT:
        """Apply a boolean equality condition to the statement."""
        if self.value is None:
            return statement
        field = self._get_instrumented_attr(model, self.field_name)
        return cast("StatementTypeT", statement.where(field == self.value))


@dataclass
class NotInCollectionFilter(InAnyFilter, Generic[T]):
    """Data required to construct a WHERE ... NOT IN (...) clause.

    This filter restricts records based on a field's absence in a collection of values.

    The filter supports both ``NOT IN`` and ``!= ANY`` operators for collection exclusion.
    Use ``prefer_any=True`` in ``append_to_statement`` to use the ``ANY`` operator.

    Parameters
    ----------
    field_name : str
        Name of the model attribute to filter on
    values : abc.Collection[T] | None
        Values for the ``NOT IN`` clause. If this is None or empty,
        the filter is not applied.

    """

    field_name: FilterFieldName
    """Field name, model attribute, or func expression."""
    values: Union[Collection[T], None]
    """Values for the ``NOT IN`` clause. If None or empty, no filter is applied."""

    def append_to_statement(
        self,
        statement: StatementTypeT,
        model: type[ModelT],
        prefer_any: bool = False,
    ) -> StatementTypeT:
        """Apply a WHERE ... NOT IN or WHERE ... != ANY(...) clause to the statement.

        Parameters
        ----------
        statement : StatementTypeT
            The SQLAlchemy statement to modify
        model : type[ModelT]
            The SQLAlchemy model class
        prefer_any : bool, optional
            If True, uses the SQLAlchemy :func:`any_` operator instead of
            :func:`notin_` for the filter condition

        Returns:
        --------
        StatementTypeT
            Modified statement with the appropriate NOT IN conditions
        """
        field = self._get_instrumented_attr(model, self.field_name)
        if not self.values:
            # If None or empty, we do not modify the statement
            return statement
        if prefer_any:
            return cast("StatementTypeT", statement.where(any_(self.values) != field))  # type: ignore[arg-type]
        return cast("StatementTypeT", statement.where(field.notin_(self.values)))


@dataclass
class NullFilter(StatementFilter):
    """Filter for NULL values (IS NULL).

    This filter creates IS NULL conditions for database fields.
    Use this to find records where a field has no value.

    Example:
        Basic NULL filtering::

            from advanced_alchemy.filters import NullFilter

            # Find records where email_verified_at is NULL
            null_filter = NullFilter("email_verified_at")
            unverified = await repo.get_many(null_filter)

        With multiple filters::

            from advanced_alchemy.filters import (
                NullFilter,
                CollectionFilter,
            )

            # Find unverified users in specific roles
            filters = [
                NullFilter("email_verified_at"),
                CollectionFilter("role", ["admin", "moderator"]),
            ]
            results = await repo.get_many(*filters)

    See Also:
        - :class:`NotNullFilter`: Filter for NOT NULL values
        - :class:`CollectionFilter`: Filter by collection membership
        - :meth:`sqlalchemy.sql.expression.ColumnOperators.is_`: IS NULL operator
    """

    field_name: FilterFieldName
    """Field name, model attribute, or func expression."""

    def append_to_statement(self, statement: StatementTypeT, model: type[ModelT]) -> StatementTypeT:
        """Apply IS NULL condition to the statement.

        Args:
            statement: The SQLAlchemy statement to modify
            model: The SQLAlchemy model class

        Returns:
            StatementTypeT: Modified statement with IS NULL condition applied
        """
        field = self._get_instrumented_attr(model, self.field_name)
        return cast("StatementTypeT", statement.where(field.is_(None)))


@dataclass
class NotNullFilter(StatementFilter):
    """Filter for NOT NULL values (IS NOT NULL).

    This filter creates IS NOT NULL conditions for database fields.
    Use this to find records where a field has a value.

    Example:
        Basic NOT NULL filtering::

            from advanced_alchemy.filters import NotNullFilter

            # Find records where email_verified_at is NOT NULL
            not_null_filter = NotNullFilter("email_verified_at")
            verified = await repo.get_many(not_null_filter)

        With multiple filters::

            from advanced_alchemy.filters import (
                NotNullFilter,
                CollectionFilter,
            )

            # Find verified users in specific roles
            filters = [
                NotNullFilter("email_verified_at"),
                CollectionFilter("role", ["admin", "moderator"]),
            ]
            results = await repo.get_many(*filters)

    See Also:
        - :class:`NullFilter`: Filter for NULL values
        - :class:`CollectionFilter`: Filter by collection membership
        - :meth:`sqlalchemy.sql.expression.ColumnOperators.is_not`: IS NOT NULL operator
    """

    field_name: FilterFieldName
    """Field name, model attribute, or func expression."""

    def append_to_statement(self, statement: StatementTypeT, model: type[ModelT]) -> StatementTypeT:
        """Apply IS NOT NULL condition to the statement.

        Args:
            statement: The SQLAlchemy statement to modify
            model: The SQLAlchemy model class

        Returns:
            StatementTypeT: Modified statement with IS NOT NULL condition applied
        """
        field = self._get_instrumented_attr(model, self.field_name)
        return cast("StatementTypeT", statement.where(field.is_not(None)))


class PaginationFilter(StatementFilter, ABC):
    """Abstract base class for pagination filters.

    Subclasses should implement pagination logic, such as limit/offset or
    cursor-based pagination.
    """


@dataclass
class LimitOffset(PaginationFilter):
    """Limit and offset pagination filter.

    Implements traditional pagination using SQL LIMIT and OFFSET clauses.
    Only applies to SELECT statements; other statement types are returned unmodified.

    Note:
        This filter only modifies SELECT statements. For other statement types
        (UPDATE, DELETE), the statement is returned unchanged.

    See Also:
        - :meth:`sqlalchemy.sql.expression.Select.limit`: SQLAlchemy LIMIT clause
        - :meth:`sqlalchemy.sql.expression.Select.offset`: SQLAlchemy OFFSET clause
    """

    limit: int
    """Maximum number of rows to return."""
    offset: int
    """Number of rows to skip before returning results."""

    def append_to_statement(self, statement: StatementTypeT, model: type[ModelT]) -> StatementTypeT:
        """Apply LIMIT/OFFSET pagination to the statement.

        Args:
            statement: The SQLAlchemy statement to modify
            model: The SQLAlchemy model class

        Returns:
            StatementTypeT: Modified statement with limit and offset applied

        Note:
            Only modifies SELECT statements. Other statement types are returned as-is.

        See Also:
            :class:`sqlalchemy.sql.expression.Select`: SQLAlchemy SELECT statement
        """
        if isinstance(statement, Select):
            statement = cast("StatementTypeT", statement.limit(self.limit).offset(self.offset))
        return statement


@dataclass
class OrderBy(StatementFilter):
    """Order by a specific field.

    Appends an ORDER BY clause to SELECT statements, sorting records by the
    specified field in ascending or descending order.

    Note:
        This filter only modifies SELECT statements. For other statement types,
        the statement is returned unchanged.

    See Also:
        - :meth:`sqlalchemy.sql.expression.Select.order_by`: SQLAlchemy ORDER BY clause
        - :meth:`sqlalchemy.sql.expression.ColumnElement.asc`: Ascending order
        - :meth:`sqlalchemy.sql.expression.ColumnElement.desc`: Descending order
    """

    field_name: FilterFieldName
    """Field name, model attribute, or func expression (e.g., ``func.random()``)."""
    sort_order: Literal["asc", "desc"] = "asc"
    """Sort direction ("asc" or "desc")."""

    def append_to_statement(self, statement: StatementTypeT, model: type[ModelT]) -> StatementTypeT:
        """Append an ORDER BY clause to the statement.

        Args:
            statement: The SQLAlchemy statement to modify
            model: The SQLAlchemy model class

        Returns:
            StatementTypeT: Modified statement with an ORDER BY clause

        Note:
            Only modifies SELECT statements. Other statement types are returned as-is.

        See Also:
            :meth:`sqlalchemy.sql.expression.Select.order_by`: SQLAlchemy ORDER BY
        """
        if isinstance(statement, Select):
            field = self._get_instrumented_attr(model, self.field_name)
            if self.sort_order == "desc":
                statement = cast("StatementTypeT", statement.order_by(field.desc()))
            else:
                statement = cast("StatementTypeT", statement.order_by(field.asc()))
        return statement


@dataclass
class SearchFilter(StatementFilter):
    """Case-sensitive or case-insensitive substring matching filter.

    Implements text search using SQL LIKE or ILIKE operators. Can search across
    multiple fields using OR conditions.

    Note:
        The search pattern automatically adds wildcards before and after the search
        value, equivalent to SQL pattern '%value%'.

    See Also:
        - :class:`.NotInSearchFilter`: Opposite filter using NOT LIKE/ILIKE
        - :meth:`sqlalchemy.sql.expression.ColumnOperators.like`: Case-sensitive LIKE
        - :meth:`sqlalchemy.sql.expression.ColumnOperators.ilike`: Case-insensitive LIKE
    """

    field_name: Union[str, set[str]]
    """Name or set of names of model attributes to search on."""
    value: str
    """Text to match within the field(s)."""
    ignore_case: Optional[bool] = False
    """Whether to use case-insensitive matching."""

    @property
    def _operator(self) -> Callable[..., ColumnElement[bool]]:
        """Return the SQL operator for combining multiple search clauses.

        Returns:
            Callable[..., ColumnElement[bool]]: The `or_` operator for OR conditions

        See Also:
            :func:`sqlalchemy.sql.expression.or_`: SQLAlchemy OR operator
        """
        return or_

    @property
    def _func(self) -> "attrgetter[Callable[[str], BinaryExpression[bool]]]":
        """Return the appropriate LIKE or ILIKE operator as a function.

        Returns:
            attrgetter: Bound method for LIKE or ILIKE operations

        See Also:
            - :meth:`sqlalchemy.sql.expression.ColumnOperators.like`: LIKE operator
            - :meth:`sqlalchemy.sql.expression.ColumnOperators.ilike`: ILIKE operator
        """
        return attrgetter("ilike" if self.ignore_case else "like")

    @property
    def normalized_field_names(self) -> set[str]:
        """Convert field_name to a set if it's a single string.

        Returns:
            set[str]: Set of field names to be searched
        """
        return {self.field_name} if isinstance(self.field_name, str) else self.field_name

    def get_search_clauses(self, model: type[ModelT]) -> list[BinaryExpression[bool]]:
        """Generate the LIKE/ILIKE clauses for all specified fields.

        Args:
            model: The SQLAlchemy model class

        Returns:
            list[BinaryExpression[bool]]: List of text matching expressions

        See Also:
            :class:`sqlalchemy.sql.expression.BinaryExpression`: SQLAlchemy expression
        """
        search_clause: list[BinaryExpression[bool]] = []
        for field_name in self.normalized_field_names:
            try:
                field = self._get_instrumented_attr(model, field_name)
                search_text = f"%{self.value}%"
                search_clause.append(self._func(field)(search_text))
            except AttributeError:
                msg = f"Skipping search for field {field_name}.  It is not found in model {model.__name__}"
                logger.debug(msg)
                continue
        return search_clause

    def append_to_statement(self, statement: StatementTypeT, model: type[ModelT]) -> StatementTypeT:
        """Append a LIKE/ILIKE clause to the statement.

        Args:
            statement: The SQLAlchemy statement to modify
            model: The SQLAlchemy model class

        Returns:
            StatementTypeT: Modified statement with text search clauses

        See Also:
            :meth:`sqlalchemy.sql.expression.Select.where`: SQLAlchemy WHERE clause
        """
        search_clauses = self.get_search_clauses(model)
        if not search_clauses:
            return statement
        where_clause = self._operator(*search_clauses)
        return cast("StatementTypeT", statement.where(where_clause))


# Regular typed dictionary for operators_map
operators_map: dict[str, Callable[[Any, Any], ColumnElement[bool]]] = {
    "eq": op.eq,
    "ne": op.ne,
    "gt": op.gt,
    "ge": op.ge,
    "lt": op.lt,
    "le": op.le,
    "in": op.in_op,
    "notin": op.notin_op,
    "between": lambda c, v: c.between(v[0], v[1]),
    "like": op.like_op,
    "ilike": op.ilike_op,
    "startswith": op.startswith_op,
    "istartswith": lambda c, v: c.ilike(v + "%"),
    "endswith": op.endswith_op,
    "iendswith": lambda c, v: c.ilike(v + "%"),
    "dateeq": lambda c, v: cast("Date", c) == v,
}

VALID_OPERATORS = set(operators_map.keys())
"""Set of valid operators that can be used in ComparisonFilter."""


@dataclass
class ComparisonFilter(StatementFilter):
    """Simple comparison filter for equality and inequality operations.

    This filter applies basic comparison operators (=, !=, >, >=, <, <=) to a field.
    It provides a generic way to perform common comparison operations.

    Args:
        field_name: Name of the model attribute to filter on
        operator: Comparison operator to use (must be one of: 'eq', 'ne', 'gt', 'ge', 'lt', 'le', 'in', 'notin', 'between', 'like', 'ilike', 'startswith', 'istartswith', 'endswith', 'iendswith', 'dateeq')
        value: Value to compare against

    Raises:
        ValueError: If an invalid operator is provided
    """

    field_name: FilterFieldName
    """Field name, model attribute, or func expression."""
    operator: str
    """Comparison operator to use (one of 'eq', 'ne', 'gt', 'ge', 'lt', 'le')."""
    value: Any
    """Value to compare against."""

    def append_to_statement(self, statement: StatementTypeT, model: type[ModelT]) -> StatementTypeT:
        """Apply a comparison operation to the statement.

        Args:
            statement: The SQLAlchemy statement to modify
            model: The SQLAlchemy model class

        Returns:
            StatementTypeT: Modified statement with the comparison condition

        Raises:
            ValueError: If an invalid operator is provided
        """
        field = self._get_instrumented_attr(model, self.field_name)
        operator_func = operators_map.get(self.operator)

        if operator_func is None:
            msg = f"Invalid operator '{self.operator}'. Must be one of: {', '.join(sorted(VALID_OPERATORS))}"
            raise ValueError(msg)

        condition = operator_func(field, self.value)
        return cast("StatementTypeT", statement.where(condition))


@dataclass
class NotInSearchFilter(SearchFilter):
    """Filter for excluding records that match a substring.

    Implements negative text search using SQL NOT LIKE or NOT ILIKE operators.
    Can exclude across multiple fields using AND conditions.

    Args:
        field_name: Name or set of names of model attributes to search on
        value: Text to exclude from the field(s)
        ignore_case: If True, uses NOT ILIKE for case-insensitive matching

    Note:
        Uses AND for multiple fields, meaning records matching any field will be excluded.

    See Also:
        - :class:`.SearchFilter`: Opposite filter using LIKE/ILIKE
        - :meth:`sqlalchemy.sql.expression.ColumnOperators.notlike`: NOT LIKE operator
        - :meth:`sqlalchemy.sql.expression.ColumnOperators.notilike`: NOT ILIKE operator
    """

    @property
    def _operator(self) -> Callable[..., ColumnElement[bool]]:
        """Return the SQL operator for combining multiple negated search clauses.

        Returns:
            Callable[..., ColumnElement[bool]]: The `and_` operator for AND conditions

        See Also:
            :func:`sqlalchemy.sql.expression.and_`: SQLAlchemy AND operator
        """
        return and_

    @property
    def _func(self) -> "attrgetter[Callable[[str], BinaryExpression[bool]]]":
        """Return the appropriate NOT LIKE or NOT ILIKE operator as a function.

        Returns:
            attrgetter: Bound method for NOT LIKE or NOT ILIKE operations

        See Also:
            - :meth:`sqlalchemy.sql.expression.ColumnOperators.notlike`: NOT LIKE
            - :meth:`sqlalchemy.sql.expression.ColumnOperators.notilike`: NOT ILIKE
        """
        return attrgetter("not_ilike" if self.ignore_case else "not_like")


@dataclass
class ExistsFilter(StatementFilter):
    """Filter for EXISTS subqueries.

    This filter creates an EXISTS condition using a list of column expressions.
    The expressions can be combined using either AND or OR logic. The filter applies
    a correlated subquery that returns only the rows from the main query that match
    the specified conditions.

    For example, if searching movies with `Movie.genre == "Action"`, only rows where
    the genre is "Action" will be returned.

    Parameters
    ----------
    values : list[ColumnElement[bool]]
        values: List of SQLAlchemy column expressions to use in the EXISTS clause
    operator : Literal["and", "or"], optional
        operator: If "and", combines conditions with AND, otherwise uses OR. Defaults to "and".

    Example:
    --------
        Basic usage with AND conditions::

            from sqlalchemy import select
            from advanced_alchemy.filters import ExistsFilter

            filter = ExistsFilter(
                values=[User.email.like("%@example.com%")],
            )
            statement = filter.append_to_statement(
                select(Organization), Organization
            )

        This will return only organizations where the user's email contains "@example.com".

        Using OR conditions::

            filter = ExistsFilter(
                values=[User.role == "admin", User.role == "owner"],
                operator="or",
            )

        This will return organizations where the user's role is either "admin" OR "owner".

    See Also:
    --------
        :class:`NotExistsFilter`: The inverse of this filter
        :func:`sqlalchemy.sql.expression.exists`: SQLAlchemy EXISTS expression
    """

    values: list[ColumnElement[bool]]
    """List of SQLAlchemy column expressions to use in the EXISTS clause."""
    operator: Literal["and", "or"] = "and"
    """If "and", combines conditions with the AND operator, otherwise uses OR."""

    @property
    def _and(self) -> Callable[..., ColumnElement[bool]]:
        """Access the SQLAlchemy `and_` operator.

        Returns:
            Callable[..., ColumnElement[bool]]: The `and_` operator for AND conditions

        See Also:
            :func:`sqlalchemy.sql.expression.and_`: SQLAlchemy AND operator
        """
        return and_

    @property
    def _or(self) -> Callable[..., ColumnElement[bool]]:
        """Access the SQLAlchemy `or_` operator.

        Returns:
            Callable[..., ColumnElement[bool]]: The `or_` operator for OR conditions

        See Also:
            :func:`sqlalchemy.sql.expression.or_`: SQLAlchemy OR operator
        """
        return or_

    def _get_combined_conditions(self) -> ColumnElement[bool]:
        """Combine the filter conditions using the specified operator.

        Returns:
            ColumnElement[bool]:
                A SQLAlchemy column expression combining all conditions with AND or OR
        """
        op = self._and if self.operator == "and" else self._or
        return op(*self.values)

    def get_exists_clause(self, model: type[ModelT]) -> ColumnElement[bool]:
        """Generate the EXISTS clause for the statement.

        Args:
            model : type[ModelT]
                The SQLAlchemy model class to correlate with

        Returns:
            ColumnElement[bool]:
                A correlated EXISTS expression for use in a WHERE clause
        """
        # Handle empty values list case
        if not self.values:
            # Use explicitly imported 'false' from sqlalchemy
            # Return SQLAlchemy FALSE expression
            return false()

        # Combine all values with AND or OR (using the operator specified in the filter)
        # This creates a single boolean expression from multiple conditions
        combined_conditions = self._get_combined_conditions()

        # Create a correlated subquery with the combined conditions
        try:
            subquery = select(1).where(combined_conditions)
            correlated_subquery = subquery.correlate(model.__table__)
            return exists(correlated_subquery)
        except Exception:  # noqa: BLE001
            return false()

    def append_to_statement(self, statement: StatementTypeT, model: type[ModelT]) -> StatementTypeT:
        """Append EXISTS condition to the statement.

        Args:
            statement : StatementTypeT
                The SQLAlchemy statement to modify
            model : type[ModelT]
            The SQLAlchemy model class

        Returns:
            StatementTypeT:
                Modified statement with EXISTS condition
        """
        # We apply the exists clause regardless of whether self.values is empty,
        # as get_exists_clause handles the empty case by returning false().
        exists_clause = self.get_exists_clause(model)
        return cast("StatementTypeT", statement.where(exists_clause))


@dataclass
class NotExistsFilter(StatementFilter):
    """Filter for NOT EXISTS subqueries.

    This filter creates a NOT EXISTS condition using a list of column expressions.
    The expressions can be combined using either AND or OR logic. The filter applies
    a correlated subquery that returns only the rows from the main query that DO NOT
    match the specified conditions.

    For example, if searching movies with `Movie.genre == "Action"`, only rows where
    the genre is NOT "Action" will be returned.

    Parameters
    ----------
    values : list[ColumnElement[bool]]
        values: List of SQLAlchemy column expressions to use in the NOT EXISTS clause
    operator : Literal["and", "or"], optional
        operator: If "and", combines conditions with AND, otherwise uses OR. Defaults to "and".

    Example:
    --------
        Basic usage with AND conditions::

            from sqlalchemy import select
            from advanced_alchemy.filters import NotExistsFilter

            filter = NotExistsFilter(
                values=[User.email.like("%@example.com%")],
            )
            statement = filter.append_to_statement(
                select(Organization), Organization
            )

        This will return only organizations where the user's email does NOT contain "@example.com".

        Using OR conditions::

            filter = NotExistsFilter(
                values=[User.role == "admin", User.role == "owner"],
                operator="or",
            )

        This will return organizations where the user's role is NEITHER "admin" NOR "owner".

    See Also:
    --------
        :class:`ExistsFilter`: The inverse of this filter
        :func:`sqlalchemy.sql.expression.not_`: SQLAlchemy NOT operator
        :func:`sqlalchemy.sql.expression.exists`: SQLAlchemy EXISTS expression
    """

    values: list[ColumnElement[bool]]
    """List of SQLAlchemy column expressions to use in the NOT EXISTS clause."""
    operator: Literal["and", "or"] = "and"
    """If "and", combines conditions with the AND operator, otherwise uses OR."""

    @property
    def _and(self) -> Callable[..., ColumnElement[bool]]:
        """Access the SQLAlchemy `and_` operator.

        Returns:
            Callable[..., ColumnElement[bool]]: The `and_` operator for AND conditions

        See Also:
            :func:`sqlalchemy.sql.expression.and_`: SQLAlchemy AND operator
        """
        return and_

    @property
    def _or(self) -> Callable[..., ColumnElement[bool]]:
        """Access the SQLAlchemy `or_` operator.

        Returns:
            Callable[..., ColumnElement[bool]]: The `or_` operator for OR conditions

        See Also:
            :func:`sqlalchemy.sql.expression.or_`: SQLAlchemy OR operator
        """
        return or_

    def _get_combined_conditions(self) -> ColumnElement[bool]:
        op = self._and if self.operator == "and" else self._or
        return op(*self.values)

    def get_exists_clause(self, model: type[ModelT]) -> ColumnElement[bool]:
        """Generate the NOT EXISTS clause for the statement.

        Args:
            model : type[ModelT]
                The SQLAlchemy model class to correlate with


        Returns:
            ColumnElement[bool]:
                A correlated NOT EXISTS expression for use in a WHERE clause
        """
        # Handle empty values list case
        if not self.values:
            # Return SQLAlchemy TRUE expression
            return true()

        # Combine conditions and create correlated subquery
        combined_conditions = self._get_combined_conditions()
        subquery = select(1).where(combined_conditions)
        correlated_subquery = subquery.correlate(model.__table__)
        return not_(exists(correlated_subquery))

    def append_to_statement(self, statement: StatementTypeT, model: type[ModelT]) -> StatementTypeT:
        """Append NOT EXISTS condition to the statement.

        Args:
            statement : StatementTypeT
                The SQLAlchemy statement to modify
            model : type[ModelT]
                The SQLAlchemy model class

        Returns:
            StatementTypeT:
                Modified statement with NOT EXISTS condition
        """
        # We apply the exists clause regardless of whether self.values is empty,
        # as get_exists_clause handles the empty case by returning true.
        exists_clause = self.get_exists_clause(model)
        return cast("StatementTypeT", statement.where(exists_clause))


@dataclass
class FilterGroup(StatementFilter):
    """A group of filters combined with a logical operator.

    This class combines multiple filters with a logical operator (AND/OR).
    It provides a way to create complex nested filter conditions.
    """

    logical_operator: Callable[..., ColumnElement[bool]]
    """Logical operator to combine the filters."""
    filters: list[StatementFilter]
    """List of filters to combine."""

    def append_to_statement(
        self,
        statement: StatementTypeT,
        model: type[ModelT],
    ) -> "StatementTypeT":
        """Apply all filters combined with the logical operator.

        Args:
            statement: The SQLAlchemy statement to modify
            model: The SQLAlchemy model class

        Returns:
            StatementTypeT: Modified statement with combined filters
        """
        if not self.filters:
            return statement

        # Create a list of expressions from each filter
        expressions = []
        for filter_obj in self.filters:
            # Each filter needs to be applied to a clean version of the statement
            # to get just its expression
            filter_statement = filter_obj.append_to_statement(select(), model)
            # Extract the whereclause from the filter's statement
            if hasattr(filter_statement, "whereclause") and filter_statement.whereclause is not None:
                expressions.append(filter_statement.whereclause)  # pyright: ignore

        if expressions:
            # Combine all expressions with the logical operator
            combined = self.logical_operator(*expressions)
            return cast("StatementTypeT", statement.where(combined))
        return statement


@dataclass
class MultiFilter(StatementFilter):
    """Apply multiple filters to a query based on a JSON/dict input.

    This filter provides a way to construct complex filter trees from
    a structured dictionary input, supporting nested logical groups and
    various filter types.
    """

    filters: dict[str, Any]
    """JSON/dict structure representing the filters."""

    # TypedDict class variables
    _filter_map: ClassVar[FilterMap] = {
        "before_after": BeforeAfter,
        "on_before_after": OnBeforeAfter,
        "boolean": BooleanFilter,
        "choices": ChoicesFilter,
        "collection": CollectionFilter,
        "not_in_collection": NotInCollectionFilter,
        "limit_offset": LimitOffset,
        "null": NullFilter,
        "not_null": NotNullFilter,
        "order_by": OrderBy,
        "search": SearchFilter,
        "not_in_search": NotInSearchFilter,
        "filter_group": FilterGroup,
        "comparison": ComparisonFilter,
        "exists": ExistsFilter,
        "not_exists": NotExistsFilter,
    }

    _logical_map: ClassVar[LogicalOperatorMap] = {
        "and_": and_,
        "or_": or_,
    }

    def append_to_statement(
        self,
        statement: StatementTypeT,
        model: type[ModelT],
    ) -> StatementTypeT:
        """Apply the filters to the statement based on the filter definitions.

        Args:
            statement: The SQLAlchemy statement to modify
            model: The SQLAlchemy model class

        Returns:
            StatementTypeT: Modified statement with all filters applied
        """
        for filter_type, conditions in self.filters.items():
            operator = self._logical_map.get(filter_type)
            if operator and isinstance(conditions, list):
                # Create filters from the conditions
                valid_filters = []
                for cond in conditions:  # pyright: ignore
                    filter_instance = self._create_filter(cond)  # pyright: ignore
                    if filter_instance is not None:
                        valid_filters.append(filter_instance)  # pyright: ignore

                # Only create a filter group if we have valid filters
                if valid_filters:
                    filter_group = FilterGroup(
                        logical_operator=operator,  # type: ignore
                        filters=valid_filters,  # pyright: ignore
                    )
                    statement = filter_group.append_to_statement(statement, model)
        return statement

    def _create_filter(self, condition: dict[str, Any]) -> Optional[StatementFilter]:
        """Create a filter instance from a condition dictionary.

        Args:
            condition: Dictionary defining a filter

        Returns:
            Optional[StatementFilter]: Filter instance if successfully created, None otherwise
        """
        # Check if condition is a nested logical group
        logical_keys = set(self._logical_map.keys())
        intersect = logical_keys.intersection(condition.keys())
        if intersect:
            # It's a nested filter group
            for key in intersect:
                operator = self._logical_map.get(key)
                if operator and isinstance(condition.get(key), list):
                    nested_filters = []
                    for cond in condition[key]:
                        filter_instance = self._create_filter(cond)
                        if filter_instance is not None:
                            nested_filters.append(filter_instance)  # pyright: ignore

                    if nested_filters:
                        return FilterGroup(logical_operator=operator, filters=nested_filters)  # type: ignore
        else:
            # Regular filter
            filter_type = condition.get("type")
            if filter_type is not None and isinstance(filter_type, str):
                filter_class = self._filter_map.get(filter_type)
                if filter_class is not None:
                    try:
                        # Create a copy of the condition without the type key
                        filter_args = {k: v for k, v in condition.items() if k != "type"}
                        return filter_class(**filter_args)  # type: ignore
                    except Exception:  # noqa: BLE001
                        return None
        return None


# Define FilterTypes using direct class references
FilterTypes: TypeAlias = Union[
    BeforeAfter,
    OnBeforeAfter,
    BooleanFilter,
    ChoicesFilter[Any],
    CollectionFilter[Any],
    LimitOffset,
    NullFilter,
    NotNullFilter,
    OrderBy,
    SearchFilter,
    NotInCollectionFilter[Any],
    NotInSearchFilter,
    ExistsFilter,
    NotExistsFilter,
    ComparisonFilter,
    MultiFilter,
    FilterGroup,
]
"""Aggregate type alias of the types supported for collection filtering."""
